﻿-- Author      : Mimma
-- Create Date : 4/3/2016 11:53:48 AM

local GBAY_MESSAGE_PREFIX		= "GBAYv1"

local GBAY_DEBUG_ENABLED		= false;

local CHAT_END					= "|r"
local COLOUR_INTRO				= "|c80F0F020"
local COLOUR_CHAT				= "|c8040A0F8"

local PARTY_CHANNEL				= "PARTY"
local RAID_CHANNEL				= "RAID"
local YELL_CHANNEL				= "YELL"
local SAY_CHANNEL				= "SAY"
local WARN_CHANNEL				= "RAID_WARNING"
local GUILD_CHANNEL				= "GUILD"
local WHISPER_CHANNEL			= "WHISPER"


--	Settings (to be made configurable)
local GBAY_CONFIG_UseGuildNotes		= 0;
local GBAY_CONFIG_MinimumBidStrategy= 1;
local GBAY_CONFIG_DKPStringLength	= 5;
local GBAY_CONFIG_MinimumDKPPenalty	= 50;	-- Minimum DKP withdrawn when doing percent DKP

--	State machine:
local STATE_NONE				= 0
local STATE_AUCTION_RUNNING		= 10
local STATE_AUCTION_PAUSED		= 20
local STATE_AUCTION_COMPLETE	= 30
local STATE_PAUSED				= 90

-- An action runs for a minimum of 20 seconds, and minimum 5 seconds after a new bid is received
local TIME_STATE_AUCTION		= 20
local TIME_STATE_EXTENSION		= 8
local GUILD_REFRESH_TIMER		= 5;		-- Check guild update every 5th second

local RAID_STATE_DISABLED		= 0
local RAID_STATE_ENABLED		= 1
-- UI Status: True = Open, False = Closed - use to prevent update of UI elements when closed.
local RaidQueueUIOpen			= false;
local TransactionUIOpen			= false;
local TransactionDetailsOpen	= false;

-- Max # of bids shown in the AuctionUI
local MAX_BIDS					= 10
-- List of valid bids: { Name, DKP }
local IncomingBidsTable			= { };

-- true if a job is already running
local JobIsRunning				= false


--[[
--	Master / Slave setup
--	--------------------
--	In order to use the "!" commands in gchat and raid chat correctly only one GBay client
--	should respond to those commands. If not, a queue command would result in multiple
--	whispers: one for each GBay client!
--
--	When a new client joins the raid, the client must issue a "WHO_IS_MASTER" request if client
--	is eligible to be Helper or Master (see pre-requisites).
--	The master (or helper) must respond back with "I_AM_MASTER" (see below), which will stop the
--	client for further requests.
--	
--	Should the Helper or Master log out / disconnect, another "master" must be found. Therefore
--	the client must issue a "WHO_IS_MASTER" (all eligible clients will do that), and each client
--	eligible for being master must respons back with their current state (which for now is SLAVE,
--	and a unique identifier (the logon time?). The client with the lowest identifier will be
--	promoted as Helper.
--	If no answer is received within <n> seconds, the client must promote himself to "HELPER"
--	(state=1).
--
--	Should a client issue a DKP command or start a bidding round, he will be promoted to Master
--	instantly. He must then send a "I_AM_MASTER" to all clients, thus demoting all other clients
--	to slave (state=0) - including an eventual Helper.
--
--	Pre-requisites to become promoted (and master):
--	* Can invite to Raid (raid leader/promoted)
--	* (can read notes? is that needed? Doubt so)
local CLIENT_STATE_SLAVE		= 0;		-- Client is only listening
local CLIENT_STATE_HELPER		= 1;		-- Client was promoted (passive master)
local CLIENT_STATE_MASTER		= 2;		-- Client issued a DKP command (active master)
local CLIENT_IDENTIFIER			= time();	-- Unique client identifier: current time.
local CLIENT_STATE				= CLIENT_STATE_SLAVE
--]]

--	For now the one issuing DKP or starting a BID round will be master. All other will be passive.
--	Helper is not supported.
local CLIENT_STATE_SLAVE		= 0;		-- Client is only listening
local CLIENT_STATE_MASTER		= 2;		-- Client issued a DKP command (active master)
local CLIENT_STATE				= CLIENT_STATE_SLAVE




-- Working variables:
local RaidState					= RAID_STATE_DISABLED
local AuctionedItemLink			= ""
local AuctionState				= STATE_NONE

-- Raid Roster: table of raid players:		{ Name, DKP, Class, Rank, Online }
local RaidRosterTable			= { }
-- Guild Roster: table of guild players:	{ Name, DKP, Class, Rank, Online }
local GuildRosterTable			= { }
local RaidRosterLazyUpdate		= false;
-- Max # of characters displayes per role in the Raid Queue UI. A caption will be inserted in top.
local MAX_RAID_QUEUE_SIZE		= 8;
-- Max # of transaction logs shown in UI (excluding Header)
local MAX_TRANSACTIONS_DISPLAYED= 18;
-- Max # of lines for class dkp displayed locally:
local MAX_CLASS_DKP_DISPLAYED	= 10;
-- Max # of lines for class dkp sent by whisper:
local MAX_CLASS_DKP_WHISPERED	= 5;

--	List of {jobname,name,dkp} tables
local JobQueue					= { }

-- Unique number for each queued raid member. Used for Sorting.
local QueueID					= 1;
-- Queued raid members: { Name, QueueID, Role , Class }
local RaidQueue					= { }

--  Transaction log: Contains a list of { timestamp, tid, author, description, state, { names, dkp } }
--	Transaction state: 0=Rolled back, 1=Active (default), 
local transactionLog			= {}
--	Current transactionID, starts out as 0 (=none).
local currentTransactionID		= 0;
local currentTransactionPage	= 1;	-- Current page shown (1=first page)
local selectedTransactionID		= nil;
--	Sync.state: 0=idle, 1=initializing, 2=synchronizing
local synchronizationState		= 0;
--	Hold RX_SYNCINIT responses when querying for a client to sync.
local syncResults				= { };

--	# of transactions displayed in /gdlog
local TRANSACTION_LIST_SIZE		= 5;
--	# of player names displayed per line when posting transaction log into guild chat
local TRANSACTION_PLAYERS_PER_LINE	= 8;
local TRANSACTION_STATE_ROLLEDBACK	= 0;
local TRANSACTION_STATE_ACTIVE		= 1;
--	Setting for transaction details screen:
local TRANSACTION_DETAILS_ROWS		= 18;
local TRANSACTION_DETAILS_COLUMNS	= 4;




local QUALITY_COLORS = {
	{0, "Poor",			{ 157,157,157 } },	--9d9d9d
	{1, "Common",		{ 255,255,255 } },	--ffffff
	{2, "Uncommon",		{  30,255,  0 } },	--1eff00
	{3, "Rare",			{   0,112,255 } },	--0070ff
	{4, "Epic",			{ 163, 53,238 } },	--a335ee
	{5, "Legendary",	{ 255,128,  0 } }	--ff8000
}

local CLASS_COLORS = {
	{ "Druid",			{ 255,125, 10 } },	--255 	125 	10		1.00 	0.49 	0.04 	#FF7D0A
	{ "Hunter",			{ 171,212,115 } },	--171 	212 	115 	0.67 	0.83 	0.45 	#ABD473 
	{ "Mage",			{ 105,204,240 } },	--105 	204 	240 	0.41 	0.80 	0.94 	#69CCF0 
	{ "Paladin",		{ 245,140,186 } },	--245 	140 	186 	0.96 	0.55 	0.73 	#F58CBA
	{ "Priest",			{ 255,255,255 } },	--255 	255 	255 	1.00 	1.00 	1.00 	#FFFFFF
	{ "Rogue",			{ 255,245,105 } },	--255 	245 	105 	1.00 	0.96 	0.41 	#FFF569
	{ "Shaman",			{ 245,140,186 } },	--245 	140 	186 	0.96 	0.55 	0.73 	#F58CBA
	{ "Warlock",		{ 148,130,201 } },	--148 	130 	201 	0.58 	0.51 	0.79 	#9482C9
	{ "Warrior",		{ 199,156,110 } }	--199 	156 	110 	0.78 	0.61 	0.43 	#C79C6E
}


--[[
	Echo a message for the local user only, including "logo"
]]
local function echo(msg)
	if msg then
		DEFAULT_CHAT_FRAME:AddMessage(COLOUR_CHAT .. msg .. CHAT_END)
	end
end

local function debugEcho(msg)
	if GBAY_DEBUG_ENABLED and msg then
		DEFAULT_CHAT_FRAME:AddMessage(COLOUR_CHAT .. "DEBUG: ".. msg .. CHAT_END)
	end
end

local function gEcho(msg)
	echo("<"..COLOUR_INTRO.."GBay"..COLOUR_CHAT.."> "..msg);
end

local function rwEcho(msg)
	SendChatMessage(msg, WARN_CHANNEL);
end

local function raidEcho(msg)
	SendChatMessage(msg, RAID_CHANNEL);
end

function guildEcho(msg)
	SendChatMessage(msg, GUILD_CHANNEL)
end

local function addonEcho(msg)
	SendAddonMessage(GBAY_MESSAGE_PREFIX, msg, "RAID")
end

local function whisper(receiver, msg)
	if receiver == UnitName("player") then
		gEcho(msg);
	else
		SendChatMessage(msg, WHISPER_CHANNEL, nil, receiver);
	end
end



--
--	SLASH COMMANDS
--

--[[
Commands:

Note that "!" commands are only executed by the client with MASTER or HELPER status to avoid
spamming multiple whispers to target.

GuildDKP            gBay /command       gbay !command       gbay /w command     Comment
------------------- ------------------- ------------------- ------------------- ------------------- 
/dkp                /gbay dkp [player]  -                   /w <o> dkp [<p>]    Current DKP status for player
/classdkp           /gbay class [class] -                   /w <o> class [<c>]  Current DKP status for class (gdclass)

/gdplus <n> <p>     /gbay +<n> <p>      -                   -                   Add <n> DKP to player <p>
/gdminus <n> <p>    /gbay -<n> <p>      -                   -                   Subtract <n> DKP from player <p>
/gdminuspct <n> <p> /gbay [-]<n>% <p>   -                   -                   Subtract <n>% DKP from player (+<n>% does not exist)

/addraid <n>        /gbay raid +<n>     -                   -                   Add <n> DKP to all players in raid (gBay: and queue)
/subtractraid <n>   /gbay raid -<n>     -                   -                   Subtract <n> DKP from players in raid (gBay: and queue)
/addrange <n>       /gbay range [+]<n>  -                   -                   Add <n> DKP to all players in range (gbay: and in queue)
/shareraid <n>      /gbay share [+]<n>  -                   -                   Share <n> DKP across all members in raid (gBay: and queue)
/sharerange <n>     -                   -                   -                   Share <n> DKP across all members in range (gBay: not supported)
/gddecay <n>        /gbay decay <n>[%]  -                   -                   Remove <n>% DKP from all guild members

-                   /gbay queue         !queue              /w <o> queue        Get queue status
-                   /gbay queue <r>     !queue <r>          /w <o> queue <r>    Queue as role <r>: <r> can be tank, melee, ranged or healer
-                   /gbay leave         !leave              /w <o> leave        Leave the raid queue

-                   /gbay <item>        -                   -                   Starts an auction for <item>
                    /startauction
-                   /gbay bid <n>       !bid <n>            /w <o> bid <n>      Bid <n> DKP on item currently being auctioned
-                   /gbay bid min       !bid min            /w <o> bid min      Bid the minimum bid on item currently being auctioned
-                   /gbay bid max       !bid max            /w <o> bid max      Bid everyting (go all out) on item currently being auctioned

-					/gbay config		-					-					Open the configuration interface
-					/gbay log			-					-					Open the transaction log interface
-					/gbay version		-					-					Check the GBay versions running
-					/gbay master		-					-					Force player to become Master (if he is raid leader or assistant)
]]


--[[
--	/gbay - main command entry handler
--]]
SLASH_GBAY_DEFAULT_COMMAND1 = "/gbay"
SlashCmdList["GBAY_DEFAULT_COMMAND"] = function(msg)
	GBAY_HandleGBayCommand(msg);
end



--[[
--	/startauction <item>
--]]
SLASH_GBAY_STARTAUCTION1 = "/startauction"
SlashCmdList["GBAY_STARTAUCTION"] = function(msg)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		
		GBAY_AddJob( function(job) GBAY_StartAuction(job[2]) end, msg, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end



function GBAY_HandleGBayCommand(msg)
	local playername = UnitName("player");
	local sign;

	--	Command: <item>
	--	Syntax: "<itemlink>"
	local _, _, itemId = string.find(msg, "item:(%d+):")
	if itemId then
		return GBAY_StartAuction(msg);
	end

		
	-- Split command into cmd (mandatory) and arg (optional)
	msg = string.lower(msg);
	local cmd, arg;
	local playerclass, playerrole;
	
	local spacepos = string.find(msg, "%s");
	if spacepos then
		_, _, cmd, arg = string.find(msg, "(%S+)%s+(.+)");
	else
		cmd = msg;
	end
	

	--	Command: config
	--	Syntax: "config"	
	if cmd == "cfg" or cmd == "config" then
		GBAY_DisplayConfigurationScreen();
		return;	
	end


	--	Command: master
	--	Syntax: "master"	
	if cmd == "master" then
		GBAY_RequestMaster(false);
		return;	
	end


	-- --	Command: sync
	-- --	Syntax: "sync"	
	-- --	Currently not working due to Timer not set up.
	-- if cmd == "sync" then
		-- if synchronizationState == 0 then
			-- GBAY_SynchronizeTransactionLog();
		-- else
			-- gEcho("A synchronization task is already running!");
		-- end
		-- return;	
	-- end
	-- 
	
	--	Command: version
	--	Syntax: "version"
	if cmd == "version" then
		if GBAY_IsInRaid(true) then
			addonEcho("TX_VERSION##");
		else
			gEcho(string.format("%s is using GBay version %s", UnitName("player"), GetAddOnMetadata("GBay", "Version")));
		end
		return;
	end
	
	
	--	Command: log
	--	Syntax: "log"
	if cmd == "log" then
		if arg and tonumber(arg) then
			selectedTransactionID = arg;
			GBAY_RefreshTransactionDetails();
			GBAY_OpenTransactionDetails();
		else	
			GBAY_OpenTransauctionUI();
		end
		return;
	end
	

	--	Command: dkp
	--	Syntax: "dkp [<playername>]"
	if cmd == "dkp" then
		return GBAY_Call_CheckPlayerDKP(arg);
	end

	--	Command: class
	--	Syntax: "class [<classname>]"
	if cmd == "class" then
		return GBAY_Call_CheckClassDKP(arg);
	end

	--	Command: queue
	--	Syntax: "queue [<role>]", "queue leave" (will fallback to "leave" below)
	if cmd == "queue" then
		if arg == "leave" then
			cmd = arg;
			arg = nil;
		else
			return GBAY_HandleQueueRequest(playername, msg);
		end
	end
	
	--	Command: leave
	--	Syntax: "leave"
	if cmd == "leave" then
		return GBAY_RemoveFromRaidQueue(playername);
	end
	
	--	Command: bid
	--	Syntax: "bid <%d>", "bid min", "bid max"
	if cmd == "bid" then
		return GBAY_HandlePlayerBid(playername, msg);
	end
	
	
	if cmd == "raid" then
		sign = string.sub(arg, 1, 1);
		--	Command: raid
		--	Syntax: "raid -<%d>"
		if sign == "-" then
			arg = string.sub(arg, 2);
			return GBAY_Call_SubtractRaidDKP(arg);
		--	Command: raid
		--	Syntax: "raid +<%d>"
		elseif sign == "+" then
			arg = string.sub(arg, 2);
			return GBAY_Call_AddRaidDKP(arg);
		else
			gEcho("DKP must be written as +999 or -999");
			return;
		end
	end

	if cmd == "range" then
		sign = string.sub(arg, 1, 1);
		--	Command: range
		--	Syntax: "range [+]<%d>"
		--	Plus is optional (default)
		if sign == "+" then
			arg = string.sub(arg, 2);
		end
		return GBAY_Call_AddRangedDKP(arg);
	end

	if cmd == "share" then
		sign = string.sub(arg, 1, 1);
		--	Command: share
		--	Syntax: "share [+]<%d>"
		--	Plus is optional (default, undocumented)
		if sign == "+" then
			arg = string.sub(arg, 2);
		end
		return GBAY_Call_ShareDKP(arg);
	end	

	--	Command: decay
	--	Syntax: "decay <%d>[%]"
	if cmd == "decay" then
		return GBAY_Call_DecayDKP(arg);		
	end


	--	Ok, the <cmd> is not a known command; we assume it is a "/gbay [+-]<dkp>" command.
	-- TODO: Add some regex check: something like "[+-]%d+[%]"
	sign = string.sub(cmd, 1, 1);


	--	Command: +
	--	Syntax: "+<%d> <playername>"
	if sign == "+" then
		local cmd = string.sub(cmd, 2);
		return GBAY_Call_AddPlayerDKP(arg, cmd);
	end
	

	if sign == "-" then
		cmd = string.sub(cmd, 2);		
		local percent = string.sub(cmd, string.len(cmd), string.len(cmd));
		if percent == "%" then
			--	Command: -
			--	Syntax: "-<%d>% <playername>" (note the percent in the end of the numeric value!)
			cmd = string.sub(cmd, 1, string.len(cmd) - 1)
			return GBAY_Call_SubtractPlayerDKPPercent(arg, cmd);
		else
			--	Command: -
			--	Syntax: "-<%d> <playername>"
			return GBAY_Call_SubtractPlayerDKP(arg, cmd);
		end
	end
	
	gEcho("Unknown command: ".. msg);
end



function GBAY_Call_CheckPlayerDKP(playername, sender)
	if playername then
		playername = GBAY_UCFirst(playername);
	else
		playername = UnitName("player");
	end		

	local dkp = GBAY_GetDKP(playername);
	if dkp then
		dkp = 1 * dkp;
		if sender then
			whisper(sender, string.format("%s have %d DKP.", playername, dkp));
		else
			echo(string.format("%s have %d DKP.", playername, dkp));
		end
	else
		if sender then
			whisper(sender, string.format("There are no DKP information for %s.", playername));
		else
			echo(string.format("There are no DKP information for %s.", playername));
		end
	end
end


function GBAY_Call_CheckClassDKP(playerclass, sender)
	if playerclass then
		playerclass = GBAY_UCFirst(playerclass);
	else
		playerclass = UnitClass("player");
	end		

	local classtable = { }
	for n=1, table.getn(GuildRosterTable), 1 do
		if GuildRosterTable[n][3] == playerclass then
			classtable[ table.getn(classtable) + 1] = GuildRosterTable[n];
		end
	end

	GBAY_SortTableDescending(classtable, 2);
	
	if sender then
		whisper(sender, string.format("Top %d DKP for %ss:", MAX_CLASS_DKP_WHISPERED, playerclass));
		for n=1, table.getn(classtable), 1 do
			if n <= MAX_CLASS_DKP_WHISPERED then
				whisper(sender, string.format("%d - %s: %d DKP", n, classtable[n][1], 1*(classtable[n][2])));
			end
		end
	else
		echo(string.format("Top %d DKP for %ss:", MAX_CLASS_DKP_DISPLAYED, playerclass));
		for n=1, table.getn(classtable), 1 do
			if n <= MAX_CLASS_DKP_DISPLAYED then
				echo(string.format("%d - %s: %d DKP", n, classtable[n][1], 1*(classtable[n][2])));
			end
		end
	end
end




function GBAY_GetQualityColor(quality)
	for n=1, table.getn(QUALITY_COLORS), 1 do
		local q = QUALITY_COLORS[n];
		if q[1] == quality then
			return q[3]
		end
	end
	
	-- Unknown quality code; can't happen! Let's just return poor quality!
	return QUALITY_COLORS[1][3];
end



--[[
	Convert a msg so first letter is uppercase, and rest as lower case.
]]
local function UCFirst(msg)
	if not msg then
		return ""
	end	

	local f = string.sub(msg, 1, 1)
	local r = string.sub(msg, 2)
	return string.upper(f) .. string.lower(r)
end




--
--	Slave / Master functions
--

--[[
--	Request MASTER status unconditionally.
--	This is to be called when a DKP command is executed or a bidding round is started.
--	Other clients are notified, and must immediately demote themselves to SLAVE.
--	Since 0.5.2
--]]
function GBAY_RequestMaster(silentmode)
	local playername = UnitName("player")
	local rank = GBAY_GetRaidRank(playername);
	--	Requires at least Assistant!
	if rank < 1 then
		if silentmode then
			debugEcho(string.format("Player %s have raid rank %d", playername, rank));
		else
			gEcho("You must be promoted before you can be a GBAY Master!");
		end
		return;
	end

	addonEcho("TX_SETMASTER#"..playername.."#");
	
	if silentmode then
		debugEcho("You are now GBAY Master.");
	else
		gEcho("You are now GBAY Master.");
	end
		
	CLIENT_STATE = CLIENT_STATE_MASTER;	
end


--[[
--	Returns TRUE if client is Master (or HELPER), FALSE if not.
--	Since 0.5.2
--]]
function GBAY_IsMaster()
	return not(CLIENT_STATE == CLIENT_STATE_SLAVE);
end


--[[
--	Handle a TX_SETMASTER message.
--	Client must set itself as slave.
--	No response is needed.
--	Since 0.5.2
--]]
function GBAY_HandleTXMaster(message, sender)
	if sender == UnitName("player") then
		return;
	end
	
	debugEcho(string.format("Master status was requested by %s.", sender));
	CLIENT_STATE = CLIENT_STATE_SLAVE;
end




--[[
--	Returns:
--	0: If player was not found or not assistant/leader
--	1: If player is assistant
--	2: If player is leader
--]]
function GBAY_GetRaidRank(playername)	
	if(GBAY_IsInRaid(true)) then	
		for n=1, GetNumRaidMembers(), 1 do
			local name, rank = GetRaidRosterInfo(n);
			if name == playername then
				return rank;
			end
		end
	end	
	return 0;
end




--
--	PROPERTIES
--

function GBAY_GetSecondCounter()
	return Seconds;
end

function GBAY_GetAuctionState()
	return AuctionState;
end

function GBAY_SetAuctionState(auctionState, seconds)
	if not seconds then
		seconds = 0;
	end
	AuctionState = auctionState;
	GBAY_setSecondCounter(seconds);
end



--
--	CONFIGURATION
--

function GBAY_DisplayConfigurationScreen()
	GBAY_OpenConfigurationUI();
end



--[[
	Start the auction, and set state to STATE_STARTING
	Parameters:
	itemLink: a Blizzard itemlink to auction.
	Since 0.0.1
]]
function GBAY_StartAuction(itemLink)
	local rank = GBAY_GetRaidRank(UnitName("player"));
	if rank < 1 then
		gEcho("You need to be Raid Assistant or Raid Leader to start auctions.");
		return;
	end


	AuctionedItemLink = itemLink;
	
	--	Poor player, not only must be handle the bidding round but he is now also handling Invites!
	GBAY_RequestMaster();
	
	-- Extract ItemId from itemLink string:
	local _, _, itemId = string.find(itemLink, "item:(%d+):")
	if not itemId then
		gEcho("Item was not found: ".. itemLink);
		return;
	end

	local itemName, _, itemQuality, _, _, _, _, _, itemTexture = GetItemInfo(itemId);	
	
	local frame = getglobal("AuctionUIFrameItem");
	if frame then
		local rgb = GBAY_GetQualityColor(itemQuality);	
		local inf = getglobal(frame:GetName().."ItemName");
		inf:SetText(itemName);
		inf:SetTextColor( (rgb[1]/255), (rgb[2]/255), (rgb[3]/255), 1);
		
		local tf = getglobal(frame:GetName().."ItemTexture");
		if tf then
			tf:SetTexture(itemTexture);
		end
	end
	
	IncomingBidsTable = { };
	GBAY_UpdateBidElements();
	GBAY_OpenAuctionUI();
	
	GBAY_SetAuctionState(STATE_AUCTION_RUNNING, TIME_STATE_AUCTION);
end



local GuildRefreshTimer = 0;
local EventTime = 0;
local TimerTick = 0;
local Secounds = 0;

function GBAY_setSecondCounter(seconds)
	Seconds = seconds;
end

function GBAY_OnTimer(elapsed)
	TimerTick = TimerTick + elapsed

	if floor(EventTime) < floor(TimerTick) then
		GBAY_CheckAuctionState();
		EventTime = TimerTick;
	end
	
	
	if floor(GuildRefreshTimer) < floor(TimerTick) then
		GuildRefreshTimer = TimerTick + GUILD_REFRESH_TIMER;
		GBAY_RequestUpdateGuildRoster();
	end
end



--[[
	The big gBay state machine.
	Since 0.0.1
]]
function GBAY_CheckAuctionState()
	local state = GBAY_GetAuctionState();
	
	--debugEcho(string.format("GBAY_CheckAuctionState called, state = %d", STATE_AUCTION_PAUSED));
	if state == STATE_NONE or state == STATE_AUCTION_PAUSED then
		return;
	end
		
	if state == STATE_AUCTION_RUNNING then
		local secs = GBAY_GetSecondCounter();
		if secs == TIME_STATE_AUCTION then
			rwEcho(string.format("Auction open for %s", AuctionedItemLink));
			rwEcho(string.format("/w %s bid [your bid]", UnitName("Player")))
		end
		
		if secs == 10 then
			rwEcho(string.format("10 seconds left - /w %s bid [your bid]", UnitName("Player")));
		end
		if secs == 3 then
			rwEcho("3 seconds left");
		end
		if secs == 2 then
			rwEcho("2 seconds left");
		end
		if secs == 1 then
			rwEcho("1 second left");
		end
		if secs < 1 then
			-- Time is up - complete the auction:
			GBAY_FinishAuction(sender, dkp);	
		end
				
		Seconds = Seconds - 1;
	end
	
	if state == STATE_COMPLETE then
		--	 We're idle
		state = STATE_NONE;
	end

	GBAY_RefreshButtonStates();
end






--[[
--	There's a message in the Guild channel - investigate that!
--]]
function GBAY_HandleGuildChatMessage(event, message, sender)
	if not message or message == "" or not string.sub(message, 1, 1) == "!" then
		return;
	end

	if GBAY_IsMaster() then
		local command = string.sub(message, 2)
		debugEcho("Master: Processing GChat command: ".. command);
		GBAY_OnChatWhisper(event, command, sender);
	end
end

--[[
--	There's a message in the Raid channel - investigate that!
--]]
function GBAY_HandleRaidChatMessage(event, message, sender)
	if not message or message == "" or not string.sub(message, 1, 1) == "!" then
		return;
	end
	
	if GBAY_IsMaster() then
		local command = string.sub(message, 2)
		debugEcho("Master: Processing RChat command: ".. command);
		GBAY_OnChatWhisper(event, command, sender);
	end
end


--[[
--	Handle incoming chat whisper.
--	Guild and Raid "!" commands are redirected here too with the "raw" command line.
--	Since 0.1.0
--]]
function GBAY_OnChatWhisper(event, message, sender)	
	if not message then
		return
	end
	
	local _, _, cmd = string.find(message, "(%a+)");	
	if not cmd then
		return
	end
	cmd = string.lower(cmd);
	
	if cmd == "bid" then
		GBAY_HandlePlayerBid(sender, message);
	elseif cmd == "queue" then
		GBAY_HandleQueueRequest(sender, message);
	end
end	



--[[
--	Handle incoming bid request.
--	Since 0.0.1
--]]
function GBAY_HandlePlayerBid(sender, message)
	local playerInfo = GBAY_GetGuildPlayerInfo(sender);
	if not playerInfo then
		whisper(sender, "You need to be in the guild for bidding!");
		return;
	end

	local availableDkp = 1 * (playerInfo[2]);
	local minimumBid = GBAY_GetMinimumBid();
	
	local _, _, dkp = string.find(message, "bid\ (%d+)")
	if not dkp then
		-- This could still be a "min" or "max" bid; try that:
		if message == "bid min" then
			dkp = minimumBid
		elseif message == "bid max" then
			dkp = availableDkp
		end
		
		if not dkp then
			-- This was not following a legal format; skip message
			return;
		end
	end
	
	local unitId = GBAY_GetUnitIDFromGroup(sender);
	if not unitId then
		-- The sender of the message was not in the raid; must be a normal whisper.
		return;
	end

	if not (AuctionState == STATE_AUCTION_RUNNING) then
		whisper(sender, "There is currently no auction running - bid was ignored.");
		return;
	end	

	dkp = 1 * dkp
	if dkp < minimumBid then
		whisper(sender, string.format("You must bid at least %s DKP - bid was ignored.", minimumBid));
		return;
	end

	if availableDkp < dkp then
		whisper(sender, string.format("You only have %d DKP - bid was ignored.", availableDkp));
		return
	end
	
	if Seconds < TIME_STATE_EXTENSION then
		Seconds = TIME_STATE_EXTENSION;
	end
	
	local bidderClass = playerInfo[3];
	local bidderRank  = playerInfo[4];
	
	rwEcho(string.format("%s is bidding %d DKP for %s", sender, dkp, AuctionedItemLink));

	GBAY_RegisterBid(sender, dkp, bidderClass, bidderRank);
	
		
	-- Checks to perform now:
	-- * Do user have enough DKP?	(Done)
	-- * Do user bid <minimum dkp>? (Done)
	--		* exception: if he goes all out he is allowed to go below minimum dkp
	-- * Is user already the highest bidder? (should we let users screw up? I personally think so!)


	-- TODO:
	-- Hide incoming whispers for local player (how?)		
end



function GBAY_RegisterBid(playername, bid, playerclass, rank)
	whisper(playername, string.format("Your bid of %d DKP has been registered.", bid) );

	IncomingBidsTable = GBAY_RenumberTable(IncomingBidsTable);
	
	IncomingBidsTable[table.getn(IncomingBidsTable) + 1] = { playername, bid, playerclass, rank };

	GBAY_SortTableDescending(IncomingBidsTable, 2);
	
	--Debug output:
	-- for n=1, table.getn(IncomingBidsTable), 1 do
		-- local cbid = IncomingBidsTable[n];
		-- local name = cbid[1];
		-- local dkp  = cbid[2];
		-- local clss = cbid[3];
		-- local rank = cbid[4];
		-- echo(string.format("%d - %s bid %d DKP, class=%s, rank=%s", n, name, dkp, clss, rank));
	-- end
 
	GBAY_UpdateBidElements();
end


function GBAY_UnregisterBid(playername, bid)
	playername = GBAY_UCFirst(playername);
	bid = 1 * bid;

	local bidInfo;
	for n=1,table.getn(IncomingBidsTable), 1 do
		bidInfo = IncomingBidsTable[n];
		if bidInfo[1] == playername and 1*(bidInfo[2]) == bid then
			table.remove(IncomingBidsTable, n);

			GBAY_UpdateBidElements();
			GBAY_ShowSelectedPlayer();
			
			return;
		end
	end
end

function GBAY_GetBidInfo(playername, bid)
	playername = GBAY_UCFirst(playername);
	bid = 1 * bid;

	local bidInfo;
	for n=1,table.getn(IncomingBidsTable), 1 do
		bidInfo = IncomingBidsTable[n];
		if bidInfo[1] == playername and 1*(bidInfo[2]) == bid then
			return bidInfo;
		end
	end

	return nil;
end

function GBAY_RenumberTable(sourcetable)
	local index = 1;
	local temptable = { };
	for key,value in ipairs(sourcetable) do
		temptable[index] = value;
		index = index + 1
	end
	return temptable;
end

function GBAY_SortTableAscending(sourcetable, index)
	local doSort = true
	while doSort do
		doSort = false
		for n=table.getn(sourcetable), 2, -1 do
			local a = sourcetable[n - 1];
			local b = sourcetable[n];
			if (a[index]) > (b[index]) then
				sourcetable[n - 1] = b;
				sourcetable[n] = a;
				doSort = true;
			end
		end
	end
	return sourcetable;
end

function GBAY_SortTableDescending(sourcetable, index)
	local doSort = true
	while doSort do
		doSort = false
		for n=1,table.getn(sourcetable) - 1, 1 do
			local a = sourcetable[n]
			local b = sourcetable[n + 1]
			if (a[index]) < (b[index]) then
				sourcetable[n] = b
				sourcetable[n + 1] = a
				doSort = true
			end
		end
	end
	return sourcetable;
end

function GBAY_CloneTable(sourcetable)
	local destinationtable = { };
	for n=1, table.getn(sourcetable), 1 do
		destinationtable[n] = sourcetable[n];
	end
	return destinationtable;
end

function GBAY_AcceptBid(playername, bid)
	if playername and bid then
		playername = GBAY_UCFirst(playername);
		bid = 1 * bid;
	
		AuctionUIFrame:Hide();
		
		rwEcho(string.format("%s sold to %s for %d DKP.", AuctionedItemLink, playername, bid));
		
		GBAY_SubtractPlayerDKP(playername, bid);		
	end
end





--
--	Player Raid Queue functions
--

--[[
--	Invite all of one role type (e.g. all Tanks)
--]]
function GBAY_InviteQueuedPlayerGroup(rolename, roleidentifier)

	StaticPopupDialogs["GBAY_POPUP_INVITE_PLAYER"] = {
		text = string.format("Do you want to invite all %s into the raid?", rolename),
		button1 = "Yes",
		button2 = "No",
		OnAccept = function() GBAY_InviteQueuedPlayerGroupNow(roleidentifier)  end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
	}
	
	StaticPopup_Show("GBAY_POPUP_INVITE_PLAYER");
end

function GBAY_InviteQueuedPlayerGroupNow(role)
	if not GBAY_IsInRaid(true) then
		return;
	end

	for n=1, table.getn(RaidQueue), 1 do
		if RaidQueue[n][3] == role then
			InviteByName(RaidQueue[n][1]);
		end
	end
end


function GBAY_InviteQueuedPlayer(playername)
	if not GBAY_IsInRaid(true) then
		return;
	end
	
	local qInfo = GBAY_GetQueuedPlayer(playername);

	if not qInfo then
		echo("Player "..playername.." is not queued!");
		return;
	end

	StaticPopupDialogs["GBAY_POPUP_INVITE_PLAYER"] = {
		text = string.format("Do you want to invite %s into the raid?", playername),
		button1 = "Yes",
		button2 = "No",
		OnAccept = function() InviteByName(playername) end,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
		preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
	}
	
	StaticPopup_Show("GBAY_POPUP_INVITE_PLAYER");
end

function GBAY_GetQueuedPlayer(playername)
	local queueInfo = nil;
	for n=1, table.getn(RaidQueue), 1 do
		queueInfo = RaidQueue[n];
		if queueInfo[1] == playername then
			return queueInfo;
		end
	end
	
	return queueInfo;
end


--[[
--	Handle incoming /queue or !queue command
--	Since: 0.1.1
--]]
function GBAY_HandleQueueRequest(sender, message)
	local _, _, queueparam = string.find(message, "queue (%a+)")
	if not queueparam then
		if sender == UnitName("player") then
			GBAY_OpenRaidQueueUI();
		else
			whisper(sender, string.format("There are currently %d people(s) in the queue.", table.getn(RaidQueue)));
		end
		return;
	end
	
	local queuetype = nil;
	
	if queueparam == "tank" or queueparam == "t" then
		queuetype = "tank";
	elseif queueparam == "melee" or queueparam == "m" then
		queuetype = "melee";
	elseif queueparam == "ranged" or queueparam == "r" then
		queuetype = "ranged";
	elseif queueparam == "healer" or queueparam == "h" then
		queuetype = "healer";
	end
	
	if queuetype then
		GBAY_AddToRaidQueue(sender, queuetype);
	else
		whisper(sender, "Type !queue <role>, where role is tank, melee, ranged or healer.");		
	end
end


--[[
--	Add a player to the raid queue
--	Since: 0.1.1
--]]
function GBAY_AddToRaidQueue(playername, playerrole)
	local playerInfo = GBAY_GetGuildPlayerInfo(playername);
	if not playerInfo then	
		whisper(playername, "You need to be in the guild to join the raid queue!");
		return;
	end
	
	local raidRoster = GBAY_GetRaidRoster();

	-- Check if player is already in the raid:
	for n=1, table.getn(raidRoster), 1 do
		if raidRoster[n][1] == playername then
			whisper(playername, "You are already in the raid.");
			return;
		end
	end

	-- Check if player is already queued:
	for n=1, table.getn(RaidQueue), 1 do
		local rq = RaidQueue[n];
		if rq[1] == playername and rq[3] == playerrole then
			whisper(playername, string.format("You are already queued as %s.", playerrole));
			return;
		end
	end

	-- Remove if already queued - that way you can change role.
	GBAY_RemoveFromRaidQueue(playername);


	-- Playername / queueId / Role / Class / Guild rank
	RaidQueue[table.getn(RaidQueue) + 1] = { playername, QueueID, playerrole, playerInfo[3], playerInfo[4] };
	QueueID = QueueID + 1;
	whisper(playername, string.format("You are now queued as %s - Queue number: %d", GBAY_UCFirst(playerrole), table.getn(RaidQueue)));

	GBAY_RefreshRaidQueue();
end



--[[
--	Remove a player from the raid queue
--]]
function GBAY_RemoveFromRaidQueue(playername)

	for n=1, table.getn(RaidQueue), 1 do
		if RaidQueue[n][1] == playername then
			RaidQueue[n] = nil;
			GBAY_RenumberTable(RaidQueue);
			return;
		end
	end

end





--
--	UI functions
--


function GBAY_OpenAuctionUI()
	GBAY_ClearSelectedPlayer();
	AuctionUIFrame:Show();
end

function GBAY_OpenRaidQueueUI()
	RaidQueueUIOpen = true;	
	GBAY_RefreshRaidQueue();
	
	RaidQueueFrame:Show();
end

function GBAY_CloseRaidQueueUI()
	RaidQueueUIOpen = false;	
	RaidQueueFrame:Hide();
end

function GBAY_OpenTransauctionUI()
	TransactionUIOpen = true;
	TransactionDetailsOpen = false;
	
	getglobal("TransactionUIFrameTableList"):Show();
	getglobal("PrevTransactionPageButton"):Show();
	getglobal("NextTransactionPageButton"):Show();
	getglobal("TransactionUIFramePlayerList"):Hide();
	getglobal("BackToTransactionLogButton"):Hide();
	getglobal("UndoTransactionButton"):Hide();
	
	GBAY_RefreshTransactionElements();
	TransactionUIFrame:Show();	
end

function GBAY_CloseTransactionUI()
	TransactionUIOpen = false;
	TransactionDetailsOpen = false;
	TransactionUIFrame:Hide();
end

function GBAY_OpenTransactionDetails()
	TransactionUIOpen = false;
	TransactionDetailsOpen = true;
	getglobal("TransactionUIFrameTableList"):Hide();
	getglobal("PrevTransactionPageButton"):Hide();
	getglobal("NextTransactionPageButton"):Hide();
	getglobal("TransactionUIFramePlayerList"):Show();
	getglobal("BackToTransactionLogButton"):Show();
	getglobal("UndoTransactionButton"):Show();
end

function GBAY_CloseTransactionDetails()
	GBAY_OpenTransauctionUI();
end

function GBAY_OpenConfigurationUI()
	ConfigurationFrame:Show();
end

function GBAY_CloseConfigurationUI()
	ConfigurationFrame:Hide();
end

function GBAY_UndoTransaction()
	-- TODO: Implement ...
end




--[[
--	Initialize all UI table elements.
--]]
function GBAY_InitializeTableElements()
	--	Initialize top <n> bids
	for n=1, MAX_BIDS, 1 do
		local entry = CreateFrame("Button", "$parentEntry"..n, AuctionUIFrameTableList, "GBAY_BidTemplate");
		entry:SetID(n);
		if n == 1 then
			entry:SetPoint("TOPLEFT", 4, -4);
		else
			entry:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
		end
	end
	
	--	Initialize Raid Queues
	for n=1, MAX_RAID_QUEUE_SIZE + 1, 1 do
		local tList = CreateFrame("Button", "$parentEntry"..n, RaidQueueFrameTankList,   "GBAY_PlayerTemplate");
		local mList = CreateFrame("Button", "$parentEntry"..n, RaidQueueFrameMeleeList,  "GBAY_PlayerTemplate");
		local rList = CreateFrame("Button", "$parentEntry"..n, RaidQueueFrameRangedList, "GBAY_PlayerTemplate");
		local hList = CreateFrame("Button", "$parentEntry"..n, RaidQueueFrameHealerList, "GBAY_PlayerTemplate");
		tList:SetID(n);
		mList:SetID(n);
		rList:SetID(n);
		hList:SetID(n);
		if(n == 1) then
			tList:SetPoint("TOPLEFT", 4, -4);
			mList:SetPoint("TOPLEFT", 4, -4);
			rList:SetPoint("TOPLEFT", 4, -4);
			hList:SetPoint("TOPLEFT", 4, -4);
		else
			tList:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
			mList:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
			rList:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
			hList:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
		end		
	end
	
	--	Initalize Transaction Log
	for n=0,MAX_TRANSACTIONS_DISPLAYED, 1 do
		local entry = CreateFrame("Button", "$parentEntry"..n, TransactionUIFrameTableList, "GBAY_LogTemplate");
		entry:SetID(n);
		if n == 0 then
			entry:SetPoint("TOPLEFT", 4, -4);
		else
			entry:SetPoint("TOP", "$parentEntry"..(n-1), "BOTTOM");
		end	
	end
	
	--	Initialize Player buttons in TransactionDetails
	local id = 1;
	for row=1, TRANSACTION_DETAILS_ROWS, 1 do
		for col=1, TRANSACTION_DETAILS_COLUMNS, 1 do
			local entry = CreateFrame("Button", "$parentEntry_"..col.."_"..row, TransactionUIFramePlayerList, "GBAY_PlayerLogTemplate");			
			entry:SetID(id);
			
			if col == 1 then
				if row == 1 then
					-- Top Left button
					entry:SetPoint("TOPLEFT", 4, -4);
				else
					-- Left button: relative to above button
					entry:SetPoint("TOP", "$parentEntry_1_"..(row-1), "BOTTOM");
				end
			else
				-- Relative to previous (left) button
				entry:SetPoint("LEFT", "$parentEntry_"..(col-1).."_"..row, "RIGHT");
			end
			
			id = id + 1;
		end
	end	
end

--	Show top <n> in bid window
function GBAY_UpdateBidElements()
	local bidder, bid, playerclass, rank;
	for n=1, MAX_BIDS, 1 do
		if table.getn(IncomingBidsTable) < n then
			bidder = "";
			bid = "";
			playerclass = "";
			rank = "";
		else
			local cbid = IncomingBidsTable[n];
			bidder = cbid[1];
			bid = string.format("%d", cbid[2]);
			playerclass = cbid[3];
			rank = cbid[4];
		end

		local color = GBAY_GetClassColorCodes(playerclass);

		local frame = getglobal("AuctionUIFrameTableListEntry"..n);
		getglobal(frame:GetName().."Bidder"):SetText(bidder);
		getglobal(frame:GetName().."Bidder"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
		getglobal(frame:GetName().."Bid"):SetText(bid);
		getglobal(frame:GetName().."Rank"):SetText(rank);

		GBAY_RefreshButtonStates();
		frame:Show();
	end
end


function GBAY_GetSelectedBid()
	local selectedBid = nil;
	
	local frame = getglobal("AuctionUIFrameSelected");
	local bidder = getglobal(frame:GetName().."Bidder"):GetText();
	local bid = getglobal(frame:GetName().."Bid"):GetText();

	if bidder and bid then
		selectedBid = { bidder, bid };
	end

	return selectedBid;
end

--[[
--	Refresh button states
--]]
function GBAY_RefreshButtonStates()
	local isAuctionRunning = (GBAY_GetAuctionState() == STATE_AUCTION_RUNNING);
	local isAuctionPaused = (GBAY_GetAuctionState() == STATE_AUCTION_PAUSED);

	local isBidderSelected = true;
	local selectedBid = GBAY_GetSelectedBid();
	if not selectedBid then
		isBidderSelected = false;
	end	

	if isBidderSelected then
		if isAuctionRunning or isAuctionPaused then
			getglobal("AcceptBidButton"):Disable();
		else
			getglobal("AcceptBidButton"):Enable();
		end		
		getglobal("CancelBidButton"):Enable();
	else
		getglobal("AcceptBidButton"):Disable();
		getglobal("CancelBidButton"):Disable();
	end
	
	if isAuctionRunning or isAuctionPaused then
		getglobal("CancelAuctionButton"):Enable();
		getglobal("RestartAuctionButton"):Enable();
		getglobal("FinishAuctionButton"):Enable();
		if isAuctionPaused then
			getglobal("PauseAuctionButton"):Enable();
			getglobal("PauseAuctionButton"):SetText("Resume Auction");
		else
			getglobal("PauseAuctionButton"):Enable();
			getglobal("PauseAuctionButton"):SetText("Pause Auction");
		end
	else
		getglobal("CancelAuctionButton"):Disable();
		getglobal("RestartAuctionButton"):Enable();
		getglobal("FinishAuctionButton"):Disable();
		getglobal("PauseAuctionButton"):Disable();
		getglobal("PauseAuctionButton"):SetText("Pause Auction");
	end	
end


--[[
--	Refresh the Raid Queue UI.
--	If a role is given, only the queue for that role is refreshed.
--	A role of nil will refresh all roles.
--]]
function GBAY_RefreshRaidQueue(role)
	-- UI not open; no need to update controls yet.
	-- This is used to allow GuildRosterUpdates to update UI when needed.
	if not RaidQueueUIOpen then
		return;
	end
	
	local raidRoster = GBAY_GetRaidRoster();
	
	local playersAlsoInRaid = { }
	
	local tQueue = { }
	local mQueue = { }
	local rQueue = { }
	local hQueue = { }
	for n=1, table.getn(RaidQueue), 1 do
		local playername = RaidQueue[n][1];
		
		local found = false;
		for f=1, table.getn(raidRoster), 1 do
			if (raidRoster[f][1] == playername) then
				found = true;
				f = table.getn(raidRoster);
			end
		end
		
		if found then
			playersAlsoInRaid[ table.getn(playersAlsoInRaid) + 1 ] = playername;
		else
			local role = RaidQueue[n][3];		
			if(role == "tank") then
				tQueue[table.getn(tQueue) + 1] = RaidQueue[n];
			elseif(role == "melee") then
				mQueue[table.getn(mQueue) + 1] = RaidQueue[n];
			elseif(role == "ranged") then
				rQueue[table.getn(rQueue) + 1] = RaidQueue[n];
			elseif(role == "healer") then
				hQueue[table.getn(hQueue) + 1] = RaidQueue[n];
			end
		end
	end
		
	--	TODO: Sort the raid tables

	local playername, playerrole, playerclass, queueid;
	
	GBAY_UpdateRaidQueueTable("Tanks",   "RaidQueueFrameTankListEntry",   tQueue);
	GBAY_UpdateRaidQueueTable("Melee",   "RaidQueueFrameMeleeListEntry",  mQueue);
	GBAY_UpdateRaidQueueTable("Ranged",  "RaidQueueFrameRangedListEntry", rQueue);
	GBAY_UpdateRaidQueueTable("Healers", "RaidQueueFrameHealerListEntry", hQueue);
	
	-- Remove players from queue who are also in Raid
	for n=1, table.getn(playersAlsoInRaid), 1 do
		GBAY_RemoveFromRaidQueue(playersAlsoInRaid[n]);
	end
	
end



function GBAY_UpdateRaidQueueTable(caption, framename, sourcetable)

	local playername, playerrole, playerclass, queueid;
	for n=0, MAX_RAID_QUEUE_SIZE, 1 do
		if n == 0 or table.getn(sourcetable) < n then
			playername	= "";
			playerrole	= "";
			queueid		= "";
			playerclass	= "";
			playerrank	= "";
		else
			playername	= sourcetable[n][1];
			queueid		= sourcetable[n][2];
			playerrole	= sourcetable[n][3];
			playerclass = sourcetable[n][4];
			playerrank	= sourcetable[n][5];
		end

		local color = { 128, 128, 128 }
		if not(playername == "") then
			local guildInfo = GBAY_GetGuildPlayerInfo(playername);
			if guildInfo then
				if guildInfo[5] == 1 then
					color = GBAY_GetClassColorCodes(playerclass);
				end
			end
		end
		
		local frame = getglobal(framename .. (n+1));		
		if n == 0 then
			local color = { 240, 200, 40 }	
			getglobal(frame:GetName().."Name"):SetText(caption);
			getglobal(frame:GetName().."Name"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
			getglobal(frame:GetName().."Rank"):SetText("Queue: ".. table.getn(sourcetable));
		else
			getglobal(frame:GetName().."Name"):SetText(playername);
			getglobal(frame:GetName().."Name"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
			getglobal(frame:GetName().."Rank"):SetText(playerrank);			
		end
		
		frame:Show();
	end
end

function GBAY_RefreshTransactionLog()
	if TransactionUIOpen then
		GBAY_RefreshTransactionElements();		
	elseif TransactionDetailsOpen then
		GBAY_RefreshTransactionDetails();	
	end
end

--[[
--	Show last <n> transactions:
--]]
function GBAY_RefreshTransactionElements()
	if not TransactionUIOpen then
		return;
	end

	local timestamp, tid, description, state, trInfo;
	local name, dkp, playerCount;
	
	local trLog = GBAY_CloneTable(transactionLog);
	GBAY_SortTableDescending(trLog, 2);
	
	local numTransactions = table.getn(trLog);
	for n=0, MAX_TRANSACTIONS_DISPLAYED, 1 do
		if n == 0 then
			timestamp = "Time";
			tid = "ID";
			description = "Command";
			trInfo = nil;
			name = "Player(s)";
			dkp = "DKP";
		else
			local index = n + ((currentTransactionPage - 1) * MAX_TRANSACTIONS_DISPLAYED);
			if numTransactions < index then
				timestamp = "";
				tid = "";
				description = "";
				trInfo = nil;
				name = "";
				dkp = "";
			else
				local tr = trLog[index];
				timestamp = tr[1];
				tid = tr[2];
				description = tr[4];
				trInfo = tr[6];
				if trInfo and table.getn(trInfo) == 1 then
					name = trInfo[1][1];
					dkp = 1 * (trInfo[1][2]);
				else
					playerCount = table.getn(trInfo);
					dkp = 0;
					for f=1, playerCount, 1 do
						dkp = dkp + 1*(trInfo[f][2]);
					end
					dkp = ceil(dkp / playerCount);
					name = string.format("(%d players)", playerCount);
				end
			end
		end

		local icon = "";
		if tonumber(dkp) then
			if dkp > 0 then
				icon = "Interface\\ICONS\\Spell_ChargePositive";
			elseif dkp < 0 then
				icon = "Interface\\ICONS\\Spell_ChargeNegative";
			end
		end

		local frame = getglobal("TransactionUIFrameTableListEntry"..n);
		getglobal(frame:GetName().."Time"):SetText(timestamp);
		getglobal(frame:GetName().."Icon"):SetTexture(icon);
		getglobal(frame:GetName().."TID"):SetText(tid);
		getglobal(frame:GetName().."Name"):SetText(name);
		getglobal(frame:GetName().."Command"):SetText(description);
		getglobal(frame:GetName().."DKP"):SetText(dkp);

		if (n > 0) and GBAY_CanReadNotes() then		
			local color = { 128, 128, 128 };
			local guildInfo = GBAY_GetGuildPlayerInfo(name);
			if guildInfo then
				color = GBAY_GetClassColorCodes(guildInfo[3]);
			end
			getglobal(frame:GetName().."Name"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
			frame:Enable();
		else
			frame:Disable();
		end

		frame:Show();
	end
	
	
	-- Refresh Transaction Buttons
	local numPages = ceil(numTransactions / MAX_TRANSACTIONS_DISPLAYED);

	if currentTransactionPage > 1 then
		getglobal("PrevTransactionPageButton"):Enable();
	else
		getglobal("PrevTransactionPageButton"):Disable();
	end
	
	if numPages > currentTransactionPage then
		getglobal("NextTransactionPageButton"):Enable();
	else
		getglobal("NextTransactionPageButton"):Disable();
	end
end

function GBAY_PreviousTransactionUIPage()
	if currentTransactionPage > 1 then
		currentTransactionPage = currentTransactionPage - 1;
	end
	GBAY_RefreshTransactionElements();
end

function GBAY_NextTransactionUIPage()
	local numTransactions = table.getn(transactionLog);
	local numPages = ceil(numTransactions / MAX_TRANSACTIONS_DISPLAYED);
	
	if numPages > currentTransactionPage then
		currentTransactionPage = currentTransactionPage + 1;
	end
	GBAY_RefreshTransactionElements();
end


function GBAY_RefreshTransactionDetails()
	selectedTransactionID = 1 * selectedTransactionID;

	local tInfo = transactionLog[selectedTransactionID];
	if not tInfo then
		return;
	end

	local playerInfo = tInfo[6];

	local totalCount = 0;	
	local totalPlayers = { };
	local found;

	local raidRoster = GBAY_GetRaidRoster();
	for n=1, table.getn(raidRoster), 1 do
		totalCount = totalCount + 1
		totalPlayers[totalCount] = { raidRoster[n][1], raidRoster[n][3] };
	end	

	for n=1, table.getn(RaidQueue), 1 do
		totalCount = totalCount + 1
		totalPlayers[totalCount] = { RaidQueue[n][1], RaidQueue[n][4] };
	end	

	totalPlayers = GBAY_SortTableAscending(totalPlayers, 1);



	local name, class, enabled;
	local row = 1;
	local col = 1;
	for n=1, TRANSACTION_DETAILS_COLUMNS * TRANSACTION_DETAILS_ROWS, 1 do
		if n > totalCount then
			name = "";
			class = "";
			enabled = false;
		else
			name = totalPlayers[n][1];
			class = totalPlayers[n][2];
			
			enabled = false;
			for f=1, table.getn(playerInfo), 1 do
				if playerInfo[f][1] == totalPlayers[n][1] then
					enabled = true;
					break;
				end
			end
		end	


		local color = { 128, 128, 128 };
		if enabled then
			color = GBAY_GetClassColorCodes(class);			
		end


		local frame = getglobal("TransactionUIFramePlayerListEntry_"..col.."_"..row);
		-- getglobal(frame:GetName().."Player"..col):SetText(name);
		-- getglobal(frame:GetName().."Player"..col):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
		getglobal(frame:GetName().."PlayerButton"):SetText(name);
		getglobal(frame:GetName().."PlayerButton"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);

		frame:Show();

		row = row + 1;
		if row > TRANSACTION_DETAILS_ROWS then
			row = 1;
			col = col + 1;
			if col > TRANSACTION_DETAILS_COLUMNS then
				break;
			end;
		end
	end
	
end


--[[
--	Accept a player bid
--	Since 0.0.3
--]]
function GBAY_AcceptSelectedPlayerBid()
	local selectedBid = GBAY_GetSelectedBid();
	if not selectedBid then
		return;
	end

	GBAY_AcceptBid(selectedBid[1], selectedBid[2]);
end


--[[
--	Cancel a player bid
--	Since 0.0.3
--]]
function GBAY_CancelSelectedPlayerBid()
	local selectedBid = GBAY_GetSelectedBid();
	if not selectedBid then
		return;
	end
	
	GBAY_UnregisterBid(selectedBid[1], selectedBid[2]);
end


--[[
--	Pause the Auction
--	Since 0.0.3
--]]
function GBAY_PauseAuction()
	local state = GBAY_GetAuctionState();	
	local secs = GBAY_GetSecondCounter();
	
	if state == STATE_AUCTION_RUNNING then
		GBAY_SetAuctionState(STATE_AUCTION_PAUSED, secs);
		rwEcho("Auction have been Paused");
	end
	
	if state == STATE_AUCTION_PAUSED then
		GBAY_SetAuctionState(STATE_AUCTION_RUNNING, secs + TIME_STATE_EXTENSION);
		rwEcho("Auction have been Resumed");
	end

	GBAY_RefreshButtonStates();
end


--[[
--	Finish the Auction
--	Since 0.0.3
--]]
function GBAY_FinishAuction()
	local state = GBAY_GetAuctionState();
	if state == STATE_AUCTION_RUNNING or state == STATE_AUCTION_PAUSED then
		rwEcho(string.format("Auction for %s is over", AuctionedItemLink));
		GBAY_SetAuctionState(STATE_AUCTION_COMPLETE);
		
		-- Check if a player was selected; if not, select highest bid:
		if table.getn(IncomingBidsTable) > 0 then
			local selectedBid = GBAY_GetSelectedBid();
			if not selectedBid then
				GBAY_ShowSelectedPlayer(IncomingBidsTable[1][1], IncomingBidsTable[1][2]);
			end
		end
	end
	
	GBAY_RefreshButtonStates();
end

--[[
--	Cancel the Auction
--	Since 0.0.3
--]]
function GBAY_CancelAuction()
	local state = GBAY_GetAuctionState();
	if state == STATE_AUCTION_RUNNING or state == STATE_AUCTION_PAUSED then
		IncomingBidsTable = { }
		GBAY_SetAuctionState(STATE_AUCTION_NONE);
		rwEcho("Auction was Cancelled");		
	end
	
	AuctionUIFrame:Hide();
end


--[[
--	Restart the Auction
--	Since 0.0.3
--]]
function GBAY_RestartAuction()
	GBAY_SetAuctionState(STATE_NONE);		
	GBAY_StartAuction(AuctionedItemLink);
end



--[[
--	Show the selected (clicked) bidder information in AuctionUI.
--	Since 0.0.2
--]]
function GBAY_ShowSelectedPlayer(playername, bid)
	local bidInfo = nil
	if playername and bid then
		bidInfo = GBAY_GetBidInfo(playername, bid);	
	end
	
	local bidder, bid, playerclass, rank;
	if not bidInfo then
		bidder = "";
		bid = "";
		playerclass = "";
		rank = "";
	else
		bidder = bidInfo[1];
		bid = string.format("%d", bidInfo[2]);
		playerclass = bidInfo[3];
		rank = bidInfo[4];
	end
	
	local color = GBAY_GetClassColorCodes(playerclass);

	local frame = getglobal("AuctionUIFrameSelected");
	getglobal(frame:GetName().."Bidder"):SetText(bidder);
	getglobal(frame:GetName().."Bidder"):SetTextColor((color[1]/255), (color[2]/255), (color[3]/255), 255);
	getglobal(frame:GetName().."Bid"):SetText(bid);
	getglobal(frame:GetName().."Rank"):SetText(rank);

	GBAY_RefreshButtonStates();
end

function GBAY_ClearSelectedPlayer()
	local frame = getglobal("AuctionUIFrameSelected");
	getglobal(frame:GetName().."Bidder"):SetText("");
	getglobal(frame:GetName().."Bid"):SetText("");
	getglobal(frame:GetName().."Rank"):SetText("");
end

function GBAY_GetClassColorCodes(classname)
	local colors = { 128,128,128 }
	classname = GBAY_UCFirst(classname);

	local cc;
	for n=1, table.getn(CLASS_COLORS), 1 do
		cc = CLASS_COLORS[n];
		if cc[1] == classname then
			return cc[2];
		end
	end

	return colors;
end


--[[
	Convert a msg so first letter is uppercase, and rest as lower case.
]]
function GBAY_UCFirst(msg)
	if not msg then
		return ""
	end	

	local f = string.sub(msg, 1, 1)
	local r = string.sub(msg, 2)
	return string.upper(f) .. string.lower(r)
end




--	****************************************************************************
--
--	DKP functions
--	Most DKP functions will also trigger a Master request.
--
--	****************************************************************************

--[[
	Get DKP belonging to a specific player.
	Returns NIL if player was not found. Players with no DKP will return 0.
]]
function GBAY_GetDKP(playername)
	local dkp = nil;
	local playerInfo = GBAY_GetGuildPlayerInfo(playername);
	if playerInfo then
		dkp = 1 * (playerInfo[2]);
	end
	
	return dkp;
 end


--[[
--	Add <dkp> DKP to <playername>
--]]
function GBAY_Call_AddPlayerDKP(playername, dkp)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_AddPlayerDKP(job[2], job[3]) end, playername, dkp )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_AddPlayerDKP(playername, dkpValue, silentmode)
	dkpValue = 1 * dkpValue;
	if GBAY_ApplyPlayerDKP(playername, dkpValue) then
		playername = GBAY_UCFirst(playername);
		rwEcho(string.format("[GBay] %d DKP was added to %s", dkpValue, playername));
		GBAY_LogSingleTransaction("+Player", playername, dkpValue);
	end
end

--[[
--	Swap players in a transaction: Remove dkp from <p> and add to <r> instead
--]]
function GBAY_Call_SwapPlayersInTransaction(transactionID, newPlayer)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_SwapPlayersInTransaction(job[2], job[3]) end, transactionID, newPlayer )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_SwapPlayersInTransaction(transactionID, newPlayer, silentmode)
	local transaction = GBAY_GetTransaction(transactionID);
	if not transaction then
		--	Transaction not found!
		return;
	end
	
	if not table.getn(transaction[6]) == 1 then
		-- Not a single-player transaction!
		return;	
	end
	
	local originalPlayer = transaction[6][1][1];
	local dkpValue = 1 * (transaction[6][1][2]);

	if GBAY_ApplyPlayerDKP(newPlayer, dkpValue) then
		newPlayer = GBAY_UCFirst(newPlayer);
		GBAY_LogSingleTransaction("+Swap", newPlayer, dkpValue);
		
		if GBAY_ApplyPlayerDKP(originalPlayer, -1 * dkpValue) then
			originalPlayer = GBAY_UCFirst(originalPlayer);
			GBAY_LogSingleTransaction("-Swap", originalPlayer, -1 * dkpValue);
			
			rwEcho(string.format("[GBay] %s was replaced with %s (%d DKP)", originalPlayer, newPlayer, dkpValue));
		end
	end
end

--[[
--	Subtract <dkp> DKP from <playername>
--]]
function GBAY_Call_SubtractPlayerDKP(playername, dkp)
	if GBAY_IsInRaid(true) and tonumber(dkp) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_SubtractPlayerDKP(job[2], job[3]) end, playername, dkp )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_SubtractPlayerDKP(playername, dkpValue, silentmode)
	dkpValue = -1 * dkpValue;
	if GBAY_ApplyPlayerDKP(playername, dkpValue) then
		playername = GBAY_UCFirst(playername);
		rwEcho(string.format("[GBay] %d DKP was subtracted from %s", abs(dkpValue), playername));
		GBAY_LogSingleTransaction("-Player", playername, dkpValue);
	end
end

function GBAY_Call_SubtractPlayerDKPPercent(playername, percent)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_SubtractPlayerDKPPercent(job[2], job[3]) end, playername, percent )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_SubtractPlayerDKPPercent(playername, percent, silentmode)
	playername = GBAY_UCFirst(playername);
	local playerInfo = GBAY_GetGuildPlayerInfo(playername);
	if playerInfo then
		percent = 1 * percent;
		local dkp = 1 * (playerInfo[2]);
		local minus = floor(dkp * percent / 100);
		if minus < GBAY_CONFIG_MinimumDKPPenalty then
			minus = GBAY_CONFIG_MinimumDKPPenalty;
		end
		
		GBAY_ApplyPlayerDKP(playername, -1 * minus, true);
		
		if not silentmode then
			rwEcho(string.format("[GBAY] %d %% (%d DKP) was subtracted from %s", percent, minus, playername));
		end

		GBAY_LogSingleTransaction("%Player", playername, -1 * abs(minus));
	else
		if not silentmode then
			gEcho(string.format("Player %s was not found", playername));
		end
	end
end

--[[
--	Add <n> DKP to all players in raid and in queue
--]]
function GBAY_Call_AddRaidDKP(dkp)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_AddRaidDKP(job[2]) end, dkp, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_AddRaidDKP(dkp, silentmode, callMethod)
	if GBAY_IsInRaid(true) then	
		dkp = 1 * dkp;
		
		if not callMethod then
			callMethod = "+Raid";
		end
		
		local tidIndex = 1
		local tidChanges = { }

		local raidRoster = GBAY_GetRaidRoster();
		for n=1, table.getn(raidRoster), 1 do
			GBAY_ApplyPlayerDKP(raidRoster[n][1], dkp);
			
			tidChanges[tidIndex] = { raidRoster[n][1], dkp };
			tidIndex = tidIndex + 1;
		end

		for n=1, table.getn(RaidQueue), 1 do
			local guildInfo = GBAY_GetGuildPlayerInfo(RaidQueue[n][1]);
			if guildInfo and guildInfo[5] == 1 then
				GBAY_ApplyPlayerDKP(RaidQueue[n][1], dkp);
				
				tidChanges[tidIndex] = { RaidQueue[n][1], dkp };
				tidIndex = tidIndex + 1;
			end
		end
		
		if not silentmode then
			rwEcho(string.format("[GBay] %d DKP was added to all players in raid", dkp));
		end
		
		GBAY_LogMultipleTransactions(callMethod, tidChanges)				
		return true;
	end
	return false;
end

--[[
--	Subtract <n> DKP from each raid and queue member.
--]]
function GBAY_Call_SubtractRaidDKP(dkp)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_SubtractRaidDKP(job[2]) end, dkp, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_SubtractRaidDKP(dkp, silentmode, callMethod)
	if GBAY_IsInRaid(true) then	
		dkp = -1 * dkp;

		if not callMethod then
			callMethod = "-Raid";
		end

		local tidIndex = 1
		local tidChanges = { }
		
		local raidRoster = GBAY_GetRaidRoster();
		for n=1, table.getn(raidRoster), 1 do
			GBAY_ApplyPlayerDKP(raidRoster[n][1], dkp);
			
			tidChanges[tidIndex] = { raidRoster[n][1], dkp };
			tidIndex = tidIndex + 1;
		end

		for n=1, table.getn(RaidQueue), 1 do
			local guildInfo = GBAY_GetGuildPlayerInfo(RaidQueue[n][1]);
			if guildInfo and guildInfo[5] == 1 then
				GBAY_ApplyPlayerDKP(RaidQueue[n][1], dkp);
				
				tidChanges[tidIndex] = { RaidQueue[n][1], dkp };
				tidIndex = tidIndex + 1;
			end
		end
		
		if not silentmode then
			rwEcho(string.format("[GBay] %d DKP was subtracted from all players in raid", abs(dkp)));
		end

		GBAY_LogMultipleTransactions(callMethod, tidChanges)
		return true;
	end
	return false;
end

--[[
--	Add <n> DKP to all in 100 yard range.
--]]
function GBAY_Call_AddRangedDKP(dkp)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_AddRangedDKP(job[2]) end, dkp, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_AddRangedDKP(dkp, silentmode)
	dkp = 1 * dkp;

	local raidUpdateCount = 0;
	local tidIndex = 1;
	local tidChanges = { };
	
	for n=1, 40, 1 do
		local unitid = "raid"..n;
		local player = UnitName(unitid);

		if player then
			if UnitIsConnected(unitid) and UnitIsVisible(unitid) then
				GBAY_ApplyPlayerDKP(player, dkp, true);
				
				tidChanges[tidIndex] = { player, dkp };
				tidIndex = tidIndex + 1;
				raidUpdateCount = raidUpdateCount + 1;
			end
		end
	end
	
	for n=1, table.getn(RaidQueue), 1 do
		local guildInfo = GBAY_GetGuildPlayerInfo(RaidQueue[n][1]);
		if guildInfo and guildInfo[5] == 1 then
			GBAY_ApplyPlayerDKP(RaidQueue[n][1], dkp);
			
			tidChanges[tidIndex] = { RaidQueue[n][1], dkp };
			tidIndex = tidIndex + 1;
		end;
	end
	
	if not silentmode then
		rwEcho(string.format("[GBAY] %d DKP has been added for %d players in range.", dkp, raidUpdateCount));
	end
	
	GBAY_LogMultipleTransactions("+Range", tidChanges)	
	return true;
end

--[[
--	Share <n> DKP to all members in raid and queue
--]]
function GBAY_Call_ShareDKP(dkp)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_RequestMaster();
		GBAY_AddJob( function(job) GBAY_ShareDKP(job[2]) end, dkp, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_ShareDKP(sharedDkp)
	if GBAY_IsInRaid(true) then	
		sharedDkp = abs(1 * sharedDkp);

		local tidIndex = 1;
		local tidChanges = { };
		
		local dkp = 0;
		local raidRoster = GBAY_GetRaidRoster();
		local count = table.getn(raidRoster);
		if count > 0 then
			dkp = ceil(sharedDkp / count);
		end
		
		if GBAY_AddRaidDKP(dkp, true, "+Share") then
			rwEcho(string.format("[GBay] %d DKP was shared (%s DKP per player)", sharedDkp, dkp));
		end
		return true;
	end
	return false;
end

--[[
--	Perform Guild Decay of <n>% DKP
--	This function requires Show Offline Members to be enabled.
--]]
function GBAY_Call_DecayDKP(percent)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_AddJob( function(job) GBAY_DecayDKP(job[2]) end, percent, "_" )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_DecayDKP(percent, silentmode)
	--	Note: arg may contain a percent sign; remove this first:
	if not tonumber(percent) then
		local percent = string.sub(percent, string.len(percent), string.len(percent));
		if percent == "%" then
			percent = string.sub(percent, 1, string.len(percent) - 1);
		end
	end
	percent = abs(1 * percent);

	--	This ensure the guild roster also contains Offline members.
	--	Otherwise offline members will not get decayed!
	if not GetGuildRosterShowOffline() == 1 then
		if not silentmode then
			gEcho("Guild Decay cancelled: You need to enable Offline Guild Members in the guild roster first.")
		end
		return false;
	end

	local tidIndex = 1;
	local tidChanges = { };

	local reducedDkp = 0;
	local playerCount = 0;

	--	Iterate over all guilded players - online or not
	local name, publicNote, officerNote
	local memberCount = GetNumGuildMembers();
	for n=1,memberCount,1 do
		name, _, _, _, _, _, publicNote, officerNote = GetGuildRosterInfo(n);
		local note = officerNote;
		if GBAY_CONFIG_UseGuildNotes == 1 then
			note = publicNote;
		end

		local _, _, dkp = string.find(note, "<(-?%d*)>");
		if dkp and tonumber(dkp)  then
			local minus = floor(dkp * percent / 100)
			tidChanges[tidIndex] = { name, (-1 * minus) }
			tidIndex = tidIndex + 1
			
			dkp = dkp - minus;
			reducedDkp = reducedDkp + minus;
			playerCount = playerCount + 1;
			note = string.gsub(note, "<(-?%d*)>", GBAY_CreateDkpString(dkp), 1);
		else
			dkp = 0;
			note = note..GBAY_CreateDkpString(dkp);
		end
		
		if GBAY_CONFIG_UseGuildNotes == 1 then
			GuildRosterSetPublicNote(n, note);
		else
			GuildRosterSetOfficerNote(n, note);
		end
		
		GBAY_UpdateLocalDKP(name, dkp);
	end
	
	if not silentmode then
		guildEcho("Guild DKP decay by "..percent.."% was performed by ".. UnitName("player") ..".")
		guildEcho("Guild DKP removed a total of "..reducedDkp.." DKP from ".. playerCount .." players.")
	end
	
	GBAY_LogMultipleTransactions("-Decay", tidChanges)
	
	return true;
end

--[[
--	Include <player> in an existing (multi-line) transaction
--]]
function GBAY_Call_IncludePlayer(transactionID, playername)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_AddJob( function(job) GBAY_IncludePlayer(job[2], job[3]) end, transactionID, playername )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_IncludePlayer(transactionID, playername, silentmode, skipApplyDkp)
	local transaction = GBAY_GetTransaction(transactionID);
	if not transaction then
		debugEcho(string.format("GBAY_IncludePlayer: Transaction not found, TID=%s", transactionID));
		return;
	end

	if not (table.getn(transaction[6]) > 0) then
		debugEcho(string.format("GBAY_IncludePlayer: There must be at least one person already, since DKP value is stored there! TID=%s", transactionID));
		return;	
	end

	-- Fetch the first valid DKP value:
	local dkpValue = 1 * (transaction[6][1][2]);	
	playername = GBAY_UCFirst(playername);


	-- Check type: It must be a Multi-line type, except for Decay:
	local trType = transaction[4];
	local validTypes = { "-Raid", "+Raid", "+Share", "+Range" }
	local found = false;
	for n=1, table.getn(validTypes), 1 do
		if validTypes[n] == trType then
			found = 1;
			break;
		end
	end
	if not found then
		debugEcho(string.format("GBAY_IncludePlayer: Invalid transaction type in TID=%s : %s", transactionID, trType));
		return;
	end
	
	-- Check player is not in the transaction already:
	for n=1, table.getn(transaction[6]), 1 do
		if transaction[6][n][1] == playername then
			debugEcho(string.format("GBAY_IncludePlayer: Player %s already exists in transaction TID=%s", playername, transactionID));
			return;
		end
	end

	transaction[6][ table.getn(transaction[6]) + 1] = { playername, dkpValue };
	debugEcho(string.format("GBAY_IncludePlayer: Player %s included in TID=%s", playername, transactionID));

	if not skipApplyDkp then	
		if GBAY_ApplyPlayerDKP(playername, dkpValue) then
			GBAY_LogIncludeExcludeTransaction("Include", playername, transactionID, dkpValue);

			echo(string.format("%s was included in transaction %d for %d DKP", playername, transactionID, dkpValue));
		end
	end
end


--[[
--	Exclude <player> from an existing (multi-line) transaction
--]]
function GBAY_Call_ExcludePlayer(transactionID, playername)
	if GBAY_IsInRaid(true) then
		RaidState = RAID_STATE_ENABLED;
		GBAY_AddJob( function(job) GBAY_ExcludePlayer(job[2], job[3]) end, transactionID, playername )
		GBAY_RequestUpdateGuildRoster();
	end
end
function GBAY_ExcludePlayer(transactionID, playername, silentmode, skipApplyDkp)
	local transaction = GBAY_GetTransaction(transactionID);
	if not transaction then
		if not transactionID then
			transactionID = "(NIL)";
		end
		debugEcho(string.format("Transaction with TID=%s was not found; cannot exclude player", transactionID));
		return;
	end

	playername = GBAY_UCFirst(playername);
	
	-- Check type: It must be a Multi-line type, except for Decay:
	local trType = transaction[4];
	local validTypes = { "-Raid", "+Raid", "+Share", "+Range" }
	local found = false;
	for n=1, table.getn(validTypes), 1 do
		if validTypes[n] == trType then
			found = 1;
			break;
		end
	end	
	if not found then
		debugEcho("Invalid transaction type: ".. trType);
		return;
	end

	-- Find player in the transaction:
	local dkpValue = nil;
	local newtable = { };
	for n=1, table.getn(transaction[6]), 1 do
		if transaction[6][n][1] == playername then
			dkpValue = transaction[6][n][2];
		else
			newtable[ table.getn(newtable) + 1] = transaction[6][n];
		end
	end
	transaction[6] = newtable;

	if not dkpValue then
		if not playername then
			playername = "(NIL)";
		end
		if not transactionID then
			transactionID = "(NIL)";
		end		
		debugEcho(string.format("The player %s was not found in the transaction with TID=%s; cannot exclude player", playername, transactionID));
		return;
	end
	
	if not skipApplyDkp then
		if GBAY_ApplyPlayerDKP(playername, -1 * dkpValue) then
			GBAY_LogIncludeExcludeTransaction("Exclude", playername, transactionID, dkpValue);			
			echo(string.format("%s was excluded from transaction %d for %d DKP", playername, transactionID, dkpValue));
		end
	end
end


--[[
--	Generic function to add(or remove) DKP from a player.
--]]
function GBAY_ApplyPlayerDKP(playername, dkpValue, silentmode)
	dkpValue = 1 * dkpValue;
	
	playername = GBAY_UCFirst(playername);
	
	local memberCount = GetNumGuildMembers()
	for n=1,memberCount,1 do
		name, _, _, _, _, _, publicNote, officerNote = GetGuildRosterInfo(n);
		if name == playername then
			local note = officerNote;
			if GBAY_CONFIG_UseGuildNotes == 1 then
				note = publicNote;
			end
		
			local _, _, dkp = string.find(note, "<(-?%d*)>");

			if dkp and tonumber(dkp)  then
				dkp = (1 * dkp) + dkpValue;
				note = string.gsub(note, "<(-?%d*)>", GBAY_CreateDkpString(dkp), 1);
			else
				dkp = dkpValue;
				note = note..GBAY_CreateDkpString(dkp);
			end
			
			if GBAY_CONFIG_UseGuildNotes == 1 then
				GuildRosterSetPublicNote(n, note);
			else
				GuildRosterSetOfficerNote(n, note);
			end
			
			GBAY_UpdateLocalDKP(name, dkp);			
			return true;
		end
   	end
   	
   	if not silentmode then
   		gEcho(string.format("%s was not found in the guild; DKP was not updated.", playername));
   	end
   	return false;
end


--[[
--	Update local stored DKP
--	Input: receiver, dkpadded
]]
function GBAY_UpdateLocalDKP(receiver, dkpAdded)

	local raidRoster = GBAY_GetRaidRoster();	--{ Name, DKP, Class, Rank, Online }
	for n=1, table.getn(raidRoster),1 do
		local player = raidRoster[n];
		local name = player[1];
		local dkp = player[2];
		local class = player[3];
		local rank = player[4];
		local online = player[5];

		if receiver == name then
			if dkp then
				dkp = dkp + dkpAdded;
			else
				dkp = dkpAdded;
			end
			
			raidRoster[n] = {name, dkp, class, rank, online};
			return;
		end
	end
end

function GBAY_CreateDkpString(dkp)
	local result;
	
	if not dkp or dkp == "" or not tonumber(dkp) then
		dkp = 0;
	end
	dkp = tonumber(dkp);
	
	local dkpLen = tonumber(GBAY_CONFIG_DKPStringLength);
	if dkpLen > 0 then
		local dkpStr = "".. abs(dkp)
		while string.len(dkpStr) < dkpLen do
			dkpStr = "0"..dkpStr;
		end
		if dkp < 0 then
			dkpStr = "-"..dkpStr;
		end				
		result = "<"..dkpStr..">";
	else
		result = "<"..dkp..">";
	end
	
	return result;
end

function GBAY_ToggleIncludePlayerInTransaction(playername)
	if not playername or not selectedTransactionID then
		return;
	end

	local transaction = GBAY_GetTransaction(selectedTransactionID);
	if not transaction then
		return;
	end

	-- See if the player is included in the transaction already:
	local currentPlayername = "";
	local includedInTransaction = false;
	local trInfo = transaction[6];
	local trSize = table.getn(trInfo);
	for n=1, trSize, 1 do
		currentPlayername = trInfo[n][1];
		if currentPlayername == playername then
			includedInTransaction = true;
			break;
		end
	end
	
	
	-- Check transaction types:
	local singlePlayerTransaction = false;
	local multiPlayerTransaction = false;
	local trType = transaction[4];

	local validTypes = { "-Player", "+Player", "%Player" }
	for n=1, table.getn(validTypes), 1 do
		if validTypes[n] == trType then
			singlePlayerTransaction = true;
			break;
		end
	end	

	local validTypes = { "-Raid", "+Raid", "+Share", "+Range" }
	for n=1, table.getn(validTypes), 1 do
		if validTypes[n] == trType then
			multiPlayerTransaction = true;
			break;
		end
	end	

	
	if includedInTransaction then
		if singlePlayerTransaction then
			--	Single-player transaction, clicked on Current player; offer to cancel transaction:
			StaticPopupDialogs["GBAY_POPUP_TRANSACTION_PLAYER"] = {
				text = "Do you want to undo this transaction?",
				button1 = "Yes",
				button2 = "No",
				OnAccept = function() GBAY_UndoTransaction(selectedTransactionID)  end,
				timeout = 0,
				whileDead = true,
				hideOnEscape = true,
				preferredIndex = 3,
			}
			StaticPopup_Show("GBAY_POPUP_TRANSACTION_PLAYER");	
		elseif multiPlayerTransaction then
			--	Multiplayer transaction, clicked on current player. Offer to exclude player:
			StaticPopupDialogs["GBAY_POPUP_TRANSACTION_PLAYER"] = {
				text = string.format("Do you want to exclude %s from this transaction?", playername),
				button1 = "Yes",
				button2 = "No",
				OnAccept = function() GBAY_ExcludePlayerFromTransaction(selectedTransactionID, playername)  end,
				timeout = 0,
				whileDead = true,
				hideOnEscape = true,
				preferredIndex = 3,
			}
			StaticPopup_Show("GBAY_POPUP_TRANSACTION_PLAYER");	
		end	
	else
		if singlePlayerTransaction then		
			--	Single-player transaction, clicked on Other player; offer to replace player:
			StaticPopupDialogs["GBAY_POPUP_TRANSACTION_PLAYER"] = {
				text = string.format("Do you want to replace %s with %s ?", currentPlayername, playername),
				button1 = "Yes",
				button2 = "No",
				OnAccept = function() GBAY_ReplacePlayerInTransaction(selectedTransactionID, currentPlayername, playername)  end,
				timeout = 0,
				whileDead = true,
				hideOnEscape = true,
				preferredIndex = 3,
			}
			StaticPopup_Show("GBAY_POPUP_TRANSACTION_PLAYER");	
		elseif multiPlayerTransaction then
			--	Multiplayer transaction, clicked on current player. Offer to exclude player:
			StaticPopupDialogs["GBAY_POPUP_TRANSACTION_PLAYER"] = {
				text = string.format("Do you want to include %s to this transaction?", playername),
				button1 = "Yes",
				button2 = "No",
				OnAccept = function() GBAY_IncludePlayerInTransaction(selectedTransactionID, playername)  end,
				timeout = 0,
				whileDead = true,
				hideOnEscape = true,
				preferredIndex = 3,
			}
			StaticPopup_Show("GBAY_POPUP_TRANSACTION_PLAYER");	
		end	
	end	
end


function GBAY_UndoTransaction(transactionID)
	--	TODO: Implement ...
end

function GBAY_ReplacePlayerInTransaction(transactionID, currentPlayer, newPlayer)
	if not currentPlayer or not newPlayer or not transactionID then
		return;
	end

	local transaction = GBAY_GetTransaction(transactionID);
	if not transaction then
		return;
	end
		
	GBAY_Call_SwapPlayersInTransaction(transactionID, newPlayer);
	GBAY_OpenTransauctionUI();
end

function GBAY_IncludePlayerInTransaction(transactionID, playername)
	if not playername or not transactionID then
		return;
	end
	GBAY_Call_IncludePlayer(transactionID, playername);
end

function GBAY_ExcludePlayerFromTransaction(transactionID, playername)
	if not playername or not transactionID then
		return;
	end
	GBAY_Call_ExcludePlayer(transactionID, playername);
end


--[[
--	Get transaction with id=<tid>
--]]
function GBAY_GetTransaction(transactionID)
	transactionID = 1 * transactionID;

	for n=1, table.getn(transactionLog), 1 do
		if (1 * transactionLog[n][2]) == transactionID then
			return transactionLog[n];
		end
	end
	return nil;
end



--[[
	Get information belonging to a specific player in the guild.
	Returns NIL if player was not found.
]]
function GBAY_GetGuildPlayerInfo(player)
	player = GBAY_UCFirst(player);

	for n=1, table.getn(GuildRosterTable), 1 do
		if GuildRosterTable[n][1] == player then
			return GuildRosterTable[n];
		end
	end
	
	return nil;

	-- local memberCount = GetNumGuildMembers()
	-- local playerInfo = nil
-- 
	-- for n=1,memberCount,1 do
		-- --	TODO: Cache the GuildRoster just like the RaidRoster
		-- local dkpValue = 0;
		-- local name, rank, _, _, playerclass, _, publicNote, officerNote = GetGuildRosterInfo(n)
		-- if name == player then
			-- echo("Found guilded player: ".. name);
			-- local note = officerNote
			-- if GBAY_CONFIG_UseGuildNotes then
				-- note = publicNote
			-- end
			-- local _, _, dkp = string.find(note, "<(-?%d*)>")
-- 
			-- if dkp and tonumber(dkp)  then
				-- dkpValue = (1 * dkp)
			-- end
			-- 
			-- playerInfo = { player, dkpValue, playerclass, rank }
		-- end
   	-- end
   	-- 
   	-- return playerInfo;
end





--
--	MinBidStrategy
--
local function strategy10Percent(dkp)
	return 1.10 * dkp;
end

function GBAY_GetHighestBid()
	local highestBid = nil;
	
	if table.getn(IncomingBidsTable) > 0 then
		highestBid = IncomingBidsTable[1];
	end

	return highestBid;
end


function GBAY_GetMinimumBid()
	local minimumBid = 100

	local highestBid = GBAY_GetHighestBid();
	if not highestBid then
		-- This is first bid = the minimum
		return minimumBid;
	end
	
	minimumBid = 1 * (highestBid[2]);

	if GBAY_CONFIG_MinimumBidStrategy == 1 then
		minimumBid = strategy10Percent(minimumBid);
	else
		-- Fallback strategy:	
		minimumBid = strategy10Percent(minimumBid);
	end

	return floor(minimumBid);
end




--
--	Raid functions
--

function GBAY_IsInRaid(silentMode)
	local result = ( GetNumRaidMembers() > 0 )
	if not silentMode and not result then
		GBAY_Echo("You must be in a raid!")
	end
	return result
end

function GBAY_CanReadNotes()
	if GBAY_CONFIG_UseGuildNotes == 1 then
		-- Guild notes can always be read; there is no WOW setting for that.
		result = true;
	else
		result = CanViewOfficerNote();
	end	
	return result
end

function GBAY_CanWriteNotes()
	if GBAY_CONFIG_UseGuildNotes == 1 then
		result = CanEditPublicNote();
	else
		result = CanViewOfficerNote() and CanEditOfficerNote();
	end
	return result
end





function GBAY_GetUnitIDFromGroup(playerName)
	playerName = UCFirst(playerName);

	if GBAY_IsInRaid(false) then
		for n=1, GetNumRaidMembers(), 1 do
			if UnitName("raid"..n) == playerName then
				return "raid"..n;
			end
		end
	else
		for n=1, GetNumPartyMembers(), 1 do
			if UnitName("party"..n) == playerName then
				return "party"..n;
			end
		end				
	end
	
	return nil;	
end




--[[
--	Button handlers:
--]]

function GBAY_OnCancelBidClick(object)
	GBAY_CancelSelectedPlayerBid();
end

function GBAY_OnPauseAuctionClick(object)
	GBAY_PauseAuction();
end

function GBAY_OnFinishAuctionClick(object)
	GBAY_FinishAuction();
end

function GBAY_OnRestartAuctionClick(object)
	GBAY_RestartAuction();
end

function GBAY_OnAcceptBidClick(object)
	GBAY_AcceptSelectedPlayerBid();
end

function GBAY_OnCancelAuctionClick(object)
	GBAY_CancelAuction();
end

function GBAY_OnBidClick(object)
	local msgID = object:GetID();
	
	local bidder = getglobal(object:GetName().."Bidder"):GetText();
	if not bidder or bidder == "" then
		return;
	end	
	local bid = 1 * (getglobal(object:GetName().."Bid"):GetText());

	GBAY_ShowSelectedPlayer(bidder, bid);
end

function GBAY_OnQueuedPlayerClick(object)
	local msgID = object:GetID();
	
	local playername = getglobal(object:GetName().."Name"):GetText();
	if not playername or playername == "" then
		return;
	end

	local playerrank = getglobal(object:GetName().."Rank"):GetText();
	if (string.len(playerrank) > 7) and (string.sub(playerrank, 1, 7) == "Queue: ") then
		if playername == "Tanks" then
			GBAY_InviteQueuedPlayerGroup(playername, "tank");
		elseif playername == "Melee" then
			GBAY_InviteQueuedPlayerGroup(playername, "melee");
		elseif playername == "Ranged" then
			GBAY_InviteQueuedPlayerGroup(playername, "ranged");
		elseif playername == "Healers" then
			GBAY_InviteQueuedPlayerGroup(playername, "healer");
		end
		
		return;
	end
	
	GBAY_InviteQueuedPlayer(playername);
end

function GBAY_OnTransactionLogClick(object)
	local msgID = object:GetID();
	selectedTransactionID = getglobal(object:GetName().."TID"):GetText();
	if not selectedTransactionID then
		return;
	end
	
	GBAY_RefreshTransactionDetails();
	GBAY_OpenTransactionDetails();
end

function GBAY_OnTransactionLogDetailPlayer(object)
	local msgID = object:GetID();
	local playername = getglobal(object:GetName().."PlayerButton"):GetText();
	
	GBAY_ToggleIncludePlayerInTransaction(playername);
end

function GBAY_OnOptionDKPStringLengthChanged(object)
	local value = getglobal(object:GetName()):GetValue();
	
	local valueString = "".. value;
	if value == 0 then
		valueString = "(No limit)";
	end
	
	getglobal(object:GetName().."Text"):SetText(string.format("DKP String Length: %s", valueString))
end

function GBAY_OnOptionMinimumDKPPenalty(object)
	local value = getglobal(object:GetName()):GetValue();
	
	local valueString = "".. value;
	if value == 0 then
		valueString = "(None)";
	end
	
	getglobal(object:GetName().."Text"):SetText(string.format("Minimum DKP penalty: %s", valueString))
end




--
--	Transaction Log handling
--
function GBAY_LogIncludeExcludeTransaction(transactioncmd, name, tid, dkp)
	local author = UnitName("Player");
	local transactions = { };
	transactions[1] = { name, dkp };
	
	local tidData = { getTimestamp(), tid, author, transactioncmd, TRANSACTION_STATE_ACTIVE, transactions };
	
	GBAY_RefreshTransactionElements();
	GBAY_BroadcastTransaction(tidData);
end



function GBAY_LogSingleTransaction(transactioncmd, name, dkp)
	local transactions = { };
	transactions[1] = { name, dkp };
	
	GBAY_LogMultipleTransactions(transactioncmd, transactions);
end


--[[
--	Add transaction and broadcast to other clients.
--]]
function GBAY_LogMultipleTransactions(transactioncmd, transactions)
	local author = UnitName("Player");
	local tid = GBAY_GetNextTransactionID();
	transactionLog[tid] = { getTimestamp(), tid, author, transactioncmd, TRANSACTION_STATE_ACTIVE, transactions };

	GBAY_RefreshTransactionElements();
	GBAY_BroadcastTransaction(transactionLog[tid]);
end

--[[
	Broadcast a transaction to other clients
]]
function GBAY_BroadcastTransaction(transaction)
	if GBAY_IsInRaid(true) then
		local timestamp = transaction[1];
		local tid = transaction[2];
		local author = transaction[3];
		local description = transaction[4];
		local transstate = transaction[5];
		local transactions = transaction[6];

		local rec, name, dkp, payload;
		for n = 1, table.getn(transactions), 1 do
			rec = transactions[n];
			name = rec[1];
			dkp = rec[2];
			--	TID plus NAME combo is unique.
			payload = timestamp .."/".. tid .."/".. author .."/".. description .."/".. transstate .."/".. name .."/".. dkp;
			addonEcho("TX_UPDATE#"..payload.."#");
		end
	end
end

function GBAY_GetNextTransactionID()
	currentTransactionID = currentTransactionID + 1;
	return currentTransactionID;
end

function GBAY_GetTimestamp()
	return date("%H:%M:%S", time());
end

--[[
	Synchronize the local transaction log with other clients.
	This is done in a two-step approach:
	- step 1:
		A TX_SYNCINIT is sent to all clients. Each clients now responds
		back (RX_SYNCINIT) with lowest and hignest TID.
		This shows how many transactions each client contains.
	- step 2:
		The client picks the response with most transactions in it,
		and will ask that client for all transactions.
		Note that step 2 will require a delay to allow all clients to
		respond back (a 2 second delay should be fine).

	Transactions are merged into existing transaction log, there is therefore
	no need to delete log fiest.
	This method should be called when GuildDKP is launched, or player enters
	a raid to make sure transactionlog is always updated.
]]
function GBAY_SynchronizeTransactionLog()
	--	This initiates step 1: send a TX_SYNCINIT to all clients.
	--	TODO: Check if we're already requesting SYNC - abort with an error if so.
	
	synchronizationState = 1	-- Step 1: Initialize
	
	addonEcho("TX_SYNCINIT##");
	
	--GBAY_AddTimer(GBAY_HandleRXSyncInitDone, 3);
end



--[[
	Timers:
	Timer structure: { Function, TimestampEnd }
]]
--[[
local GeneralTimers = {}
local GeneralTimerTick = 0

function GBAY_OnTimer(elapsed)
	GeneralTimerTick = GeneralTimerTick + elapsed

	for n=1,table.getn(GeneralTimers),1 do
		local timer = GeneralTimers[n];
		if GeneralTimerTick > timer[2] then
			GeneralTimers[n] = nil;
			timer[1]();
		end
	end
end

function GBAY_AddTimer( method, duration )
	GeneralTimers[table.getn(GeneralTimers) + 1] = { method, GeneralTimerTick + duration }
end
--]]




--
--	Message Handling
--

--[[
	Respond to a TX_VERSION command.
	Input:
		msg is the raw message
		sender is the name of the message sender.
	We should whisper this guy back with our current version number.
	We therefore generate a response back (RX) in raid with the syntax:
	GuildDKP:<sender (which is actually the receiver!)>:<version number>
]]
local function GBAY_HandleTXVersion(message, sender)
	addonEcho(string.format("RX_VERSION#%s#%s", GetAddOnMetadata("GBay", "Version"), sender));
end

--[[
--	A version response (RX) was received. The version information is displayed locally.
--]]
local function GBAY_HandleRXVersion(message, sender)
	gEcho(string.format("%s is using GBay version %s", sender, message));
end


--[[
--	TX_UPDATE: A transaction was broadcasted. Add transaction details to transactions list.
--]]
local function GBAY_HandleTXUpdate(message, sender)
	--	Message was from SELF, no need to update transactions since I made them already!
	if (sender == UnitName("player")) then
		return
	end

	local _, _, timestamp, tid, author, description, transstatus, name, dkp = string.find(message, "([^/]*)/([0-9]*)/([^/]*)/([^/]*)/([0-9]*)/([^/]*)/([^/]*)")

	tid = tonumber(tid);


	-- "Include" and "Exclude" does not have transactions, and should not be stored in the transaction log.
	-- Therefore we "catch" them here and perform the include/exclude operation as needed.
	-- These operations only update the local transaction log; no DKP is actually moved since this was
	-- already done by the master.
	if description == "Include" then
		GBAY_IncludePlayer(tid, name, true, true);
		GBAY_RefreshTransactionElements();	
		return;
	elseif description == "Exclude" then
		GBAY_ExcludePlayer(tid, name, true, true);		
		GBAY_RefreshTransactionElements();
		return;
	end
	
	local transaction = transactionLog[tid];
	if not transaction then
		transaction = { timestamp, tid, author, description, transstatus, { } };
	end
	
	--	List of transaction lines contained in this transaction ("name=dkp" entries)
	local transactions = transaction[6];
	transactions[table.getn(transactions) + 1] = { name, dkp }
	transaction[6] = transactions;

	transactionLog[tid] = transaction;

	-- Make sure to update next transactionid
	if currentTransactionID < tid then
		currentTransactionID = tid;
	end

	GBAY_RefreshTransactionLog();
end



--	Clients must return the highest transaction ID they own in RX_SYNCINIT
function GBAY_HandleTXSyncInit(message, sender)
	--	Message was from SELF, no need to return RX_SYNCINIT
	if (sender == UnitName("player")) then
		return
	end

	syncResults = { }
	addonEcho("RX_SYNCINIT#"..currentTransactionID.."#"..sender)
end


--Handle RX_SYNCINIT responses from clients
function GBAY_HandleRXSyncInit(message, sender)
	--	Check we are still in TX_SYNCINIT state
	if not (synchronizationState == 1) then
		return
	end

	local maxTid = tonumber(message);
	local syncIndex = table.getn(syncResults) + 1;
	
	syncResults[syncIndex] = { sender, message };
end


--	This is called by the timer when responses are no longer accepted
function GBAY_HandleRXSyncInitDone()
	synchronizationState = 2
	local maxTid = 0
	local maxName = ""

	for n = 1, table.getn(syncResults), 1 do
		local res = syncResults[n]
		local tid = tonumber(res[2])
		if(tid > maxTid) then
			maxTid = tid
			maxName = res[1]
		end
	end

	--	No transactions was found, nothing to sync.
	if maxTid == 0 then
		synchronizationState = 0
	end

	if maxTid > currentTransactionID then
		currentTransactionID = maxTid
	end

	--	Now request transaction synchronization from selected target
	addonEcho("TX_SYNCTRAC##"..maxName);
end


--	Client is requested to sync transaction log with <sender>
function GBAY_HandleTXSyncTransaction(message, sender)
	--	Iterate over transactions
	for n = 1, table.getn(transactionLog), 1 do
		local rec = transactionLog[n]
		local timestamp = rec[1]
		local tid = rec[2]
		local author = rec[3]
		local desc = rec[4]
		local state = rec[5]
		local tidChanges = rec[6]

		--	Iterate over transaction lines
		for f = 1, table.getn(tidChanges), 1 do
			local change = tidChanges[f]
			local name = change[1]
			local dkp = change[2]
			
			local response = timestamp.."/"..tid.."/"..author.."/"..desc.."/"..state.."/"..name.."/"..dkp;
			
			addonEcho("RX_SYNCTRAC#"..response.."#"..sender);
		end
	end
	
	--	Last, send an EOF to signal all transactions were sent.
	addonEcho("RX_SYNCTRAC#EOF#"..sender);
end


--	Received a sync'ed transaction - merge this with existing transaction log.
function GBAY_HandleRXSyncTransaction(message, sender)
	if message == "EOF" then
		synchronizationState = 0;
		return;
	end

	local _, _, timestamp, tid, author, description, transstatus, name, dkp = string.find(message, "([^/]*)/([0-9]*)/([^/]*)/([^/]*)/([0-9]*)/([^/]*)/([^/]*)")

	tid = tonumber(tid);

	local transaction = transactionLog[tid];
	if not transaction then
		transaction = { timestamp, tid, author, description, transstatus, {} };
	end

	local transactions = transaction[6];
	local tracCount = table.getn(transactions);

	--	Check if this transaction line does already exist in transaction
	for f = 1, tracCount, 1 do
		local trac = transactions[f];
		local currentName = trac[1];
		local currentDkp = trac[2];

		--	This entry already exists - no need to process further.
		if currentName == name then
			return;
		end
	end

	--	If we end here, then the transaction does not exist in our transaction log.
	--	Create entry:
	transactions[tracCount + 1] = { name, dkp };
	transaction[6] = transactions;
	transactionLog[tid] = transaction;
end






--
--	Job Handling
--
function GBAY_AddJob( method, arg1, arg2 )
	JobQueue[table.getn(JobQueue) + 1] = { method, arg1, arg2 }
end

function GBAY_GetNextJob()
	local job
	local cnt = table.getn(JobQueue)
	
	if cnt > 0 then
		job = JobQueue[1]
		for n=2,cnt,1 do
			JobQueue[n-1] = JobQueue[n]			
		end
		JobQueue[cnt] = nil
	end

	return job
end




--
--	Guild and Raid Roster Event Handling
--
--	RaidRoster will update internal raid roster table.
--	GuildRoster will update guild roster table and check job queue
--

function GBAY_RequestUpdateGuildRoster()
	GuildRoster()
end

function GBAY_OnGuildRosterUpdate()
	GBAY_RefreshGuildRoster();

	if GBAY_CanReadNotes() then
		if not JobIsRunning then	
			JobIsRunning = true
			
			local job = GBAY_GetNextJob()
			while job do
				job[1](job)				
				job = GBAY_GetNextJob()
			end
			
			if GBAY_IsInRaid(true) then
				GBAY_RefreshRaidRoster()
			end
 
			JobIsRunning = false
		end
	end
	
	GBAY_RefreshRaidQueue();
	GBAY_RefreshTransactionLog();
end


--[[
	Update the guild roster status cache: members and DKP.
	Used to display DKP values for non-raiding members
	(/gdclass and /gdstat)
]]
function GBAY_RefreshGuildRoster()
	--echo("Refreshing GuildRoster");

	GuildRosterTable = { }
	
	if not GBAY_CanReadNotes() then
		return;
	end

	local memberCount = GetNumGuildMembers();
	local note
	for n=1,memberCount,1 do
		local name, rank, _, _, class, _, publicnote, officernote, online = GetGuildRosterInfo(n)

		if GBAY_CONFIG_UseGuildNotes == 1 then		
			note = publicnote
		else
			note = officernote
		end
		
		if not note or note == "" then
			note = "<0>";
		end
		
		if not online then
			online = 0;
		end
		
		local _, _, dkp = string.find(note, "<(-?%d*)>")
		if not dkp then
			dkp = 0;
		end
		
		--echo(string.format("Added %s (%s)", name, online));
		
		GuildRosterTable[n] = { name, (1 * dkp), class, rank, online }
	end
end


function GBAY_OnRaidRosterUpdate(event, arg1, arg2, arg3, arg4, arg5)
	RaidRosterLazyUpdate = true;

	GBAY_RefreshRaidQueue();
	GBAY_RefreshTransactionLog();
	
	if GBAY_IsInRaid(true) then
		GBAY_SynchronizeTransactionLog();
	else
		transactionLog = { };
	end	
end

function GBAY_GetRaidRoster()
	if RaidRosterLazyUpdate then
		GBAY_RefreshRaidRoster();
	end
	return RaidRosterTable;
end

--[[
	Re-read the raid status and namely the DKP values.
	Should be called after each roster update.
]]
function GBAY_RefreshRaidRoster()
	local playerCount = GetNumRaidMembers()
	
	if playerCount then
		RaidRosterTable = { }
		local index = 1
		local memberCount = table.getn(GuildRosterTable);
		for n=1,playerCount,1 do
			local name, _, _, _, class = GetRaidRosterInfo(n);

			for m=1,memberCount,1 do
				local info = GuildRosterTable[m]
				if name == info[1] then
					RaidRosterTable[index] = info;
					index = index + 1
				end
			end
		end
	end
	
	RaidRosterLazyUpdate = false;
	
	for n=1,table.getn(RaidRosterTable), 1 do
		local rr = RaidRosterTable[n];
--		echo("RaidRosterUpdate:");
--		echo(string.format("Name=%s, DKP=%d, Class=%s, Rank=%s", rr[1], rr[2], rr[3], rr[4]));
	end
end



function GBAY_OnChatMessage(event, message, sender, language, channelString)
	--("message", "sender", "language", "channelString", "target", "flags", unknown, channelNumber, "channelName", unknown, counter, "guid")
	echo(string.format("Incoming msg, channel=%s, message=%s", channelString, message));

	if sender and message then
		if channelString == GUILD_CHANNEL then
			echo("Guild message found");
			GBAY_HandleGuildChatMessage(sender, message);
		elseif channelString == RAID_CHANNEL then
			echo("Raid message found");
			GBAY_HandleRaidChatMessage(sender, message);
		else
			echo(string.format("Other channel, channel=%s, message=%s", channelString, message));
		end
	end
end


function GBAY_OnChatMsgAddon(event, prefix, msg, channel, sender)

	if (prefix == GBAY_MESSAGE_PREFIX) or (prefix == "GuildDKPv1") then
	
		--	Split incoming message in Command, Payload (message) and Recipient
		local _, _, cmd, message, recipient = string.find(msg, "([^#]*)#([^#]*)#([^#]*)")

		if not cmd then
			return	-- cmd is mandatory, remaining parameters are optionel.
		end

		--	Ignore message if it is not for me. Receipient can be blank, which means it is for everyone.
		if not (recipient == "") then
			if not (recipient == UnitName("player")) then
				return
			end
		end
		
		if not message then
			message = ""
		end
	
		if cmd == "TX_VERSION" then
			GBAY_HandleTXVersion(message, sender)
		elseif cmd == "RX_VERSION" then
			GBAY_HandleRXVersion(message, sender)
		elseif cmd == "TX_UPDATE" then
			GBAY_HandleTXUpdate(message, sender)
		elseif cmd == "TX_SYNCINIT" then
			GBAY_HandleTXSyncInit(message, sender)
		elseif cmd == "RX_SYNCINIT" then
			GBAY_HandleRXSyncInit(message, sender)
		elseif cmd == "TX_SYNCTRAC" then
			GBAY_HandleTXSyncTransaction(message, sender)
		elseif cmd == "RX_SYNCTRAC" then
			GBAY_HandleRXSyncTransaction(message, sender)
		elseif cmd == "TX_SETMASTER" then
			GBAY_HandleTXMaster(message, sender)		
		else
			gEcho("Unknown command, raw msg="..msg)
		end
	end
end


--
--	Events
--
function GBAY_OnEvent(event, arg1, arg2, arg3, arg4, arg5)
	if (event == "CHAT_MSG_GUILD") then
		GBAY_HandleGuildChatMessage(event, arg1, arg2, arg3, arg4, arg5);
	elseif (event == "CHAT_MSG_RAID" or event == "CHAT_MSG_RAID_LEADER") then
		GBAY_HandleRaidChatMessage(event, arg1, arg2, arg3, arg4, arg5);
	elseif (event == "CHAT_MSG_WHISPER") then
		GBAY_OnChatWhisper(event, arg1, arg2, arg3, arg4, arg5);
	elseif (event == "CHAT_MSG_ADDON") then
		GBAY_OnChatMsgAddon(event, arg1, arg2, arg3, arg4, arg5)
	elseif (event == "GUILD_ROSTER_UPDATE") then
		GBAY_OnGuildRosterUpdate(event, arg1, arg2, arg3, arg4, arg5)
	elseif (event == "RAID_ROSTER_UPDATE") then
		GBAY_OnRaidRosterUpdate(event, arg1, arg2, arg3, arg4, arg5)
	end
end

function GBAY_OnLoad()
	gEcho(string.format("Loot Distribution Addon version %s by %s", GetAddOnMetadata("GBay", "Version"), GetAddOnMetadata("GBay", "Author")));
    
    this:RegisterEvent("GUILD_ROSTER_UPDATE")
    this:RegisterEvent("RAID_ROSTER_UPDATE");
	this:RegisterEvent("CHAT_MSG_GUILD");
	this:RegisterEvent("CHAT_MSG_RAID");
	this:RegisterEvent("CHAT_MSG_RAID_LEADER");
    this:RegisterEvent("CHAT_MSG_WHISPER");
    this:RegisterEvent("CHAT_MSG_ADDON")
    
    
    GBAY_SetAuctionState(STATE_NONE);
    GBAY_RefreshRaidRoster();
	GBAY_InitializeTableElements(); 
	
	GBAY_RequestUpdateGuildRoster()
	
	if GBAY_IsInRaid(true) then	
		GBAY_SynchronizeTransactionLog();
	end
end

